from __future__ import annotations

from queue import Empty, Queue
from typing import Callable, Literal, Optional

from prettyqt import core, gui, widgets
from prettyqt.qt import QtWidgets


CURSOR_MARGIN_TOP = 10
CURSOR_MARGIN_BOTTOM = 50
CURSOR_MARGIN_LEFT = 10
CURSOR_MARGIN_RIGHT = 10

DEFAULT_NOTIFICATION_STYLES = """
    Notification {
        font-size: 16px;
        padding: 0px;
        margin: 0px;
        border-radius: 6px;
    }
    Notification #message{
        color: #FFFFFF;
        padding: 0px;
        margin: 0px;
        width: 100%;
    }
    Notification #closeButton{
        color: #FFFFFF;
        padding: 0px;
        margin: 0px;
    }
    Notification#primary {
        background-color: #337ab7;
        border-color: #2e6da4;
    }
    Notification#success {
        background-color: #5cb85c;
        border-color: #4cae4c;
    }
    Notification#info {
        background-color: #5bc0de;
        border-color: #46b8da;
    }
    Notification#warning {
        background-color: #f0ad4e;
        border-color: #eea236;
    }
    Notification#danger {
        background-color: #d9534f;
        border-color: #d43f3a;
    }
    """

CategoryStr = Literal["primary", "success", "info", "warning", "danger"]
FadeOutValue = Literal[None, "fade_out"]
FadeInValue = Literal[None, "fade_in"]


class MessageLabel(widgets.Label):
    """Subclass of QLabel, which reimplements the resizeEvent() function.

    This is necessary because otherwise the notifications take up too much vertical
    space when texts they display become longer. This is because normally the height
    of a notification is calculated as the minimum height necessary for the text
    when the widget is horizontally resized to its minimum.
    """

    def resizeEvent(self, event):
        super().resizeEvent(event)
        if (
            self.wordWrap()
            and self.sizePolicy().verticalPolicy() == widgets.SizePolicy.Minimum
        ):
            new_height = self.heightForWidth(self.width())
            if new_height < 1:
                return
            self.setMaximumHeight(new_height)


class Notification(widgets.Widget):
    """Class representing a single notification."""

    close_clicked = core.Signal()

    def __init__(
        self,
        message: str,
        category: CategoryStr,
        timeout=None,
        autohide: bool = False,
        buttontext: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        # Store instance variables
        self.message = message
        self.category = category
        self.timeout = timeout
        self.autohide = autohide

        # Set Object name for reference
        self.setObjectName(category)
        self.set_layout("horizontal", margin=0)

        # Create a message area
        message_area = widgets.BoxLayout("horizontal")
        message_area.set_margin(0)

        # Create the layout
        self.message_display = MessageLabel()
        self.message_display.setObjectName("message")
        self.message_display.set_size_policy("minimum", "minimum")
        self.message_display.setWordWrap(True)

        # Create a button that can close notifications
        if not buttontext:
            close_button = widgets.PushButton("\u2715")
        else:
            close_button = widgets.PushButton(buttontext)
            close_button.setStyleSheet("text-decoration: underline;")
        close_button.set_size_policy("fixed", "fixed")
        close_button.setFlat(True)
        close_button.setObjectName("closeButton")
        close_button.clicked.connect(self.close_clicked)

        # Add everything together
        message_area.addWidget(self.message_display)
        # message_area.addStretch(1)
        message_area.addWidget(close_button)
        self.layout().addLayout(message_area)

        # Initialize some variables
        # self.setStyle(category)
        self.setVisible(False)

        # Flag that is set if notification is being removed. This can be used to
        # make sure that even though the notification has not been really removed
        # yet (because it is for example in a fade out animation), it is in the
        # process of being removed
        self.is_being_removed = False
        self.is_fading_in = False
        self.opacity_effect = widgets.GraphicsOpacityEffect(self)

        # Fade in animation
        self.fade_in_anim = core.PropertyAnimation()
        self.fade_in_anim.apply_to(self.opacity_effect, "opacity")
        self.fade_in_anim.set_range(0.0, 1.0)

        # Fade out animation
        self.fade_out_anim = core.PropertyAnimation()
        self.fade_out_anim.apply_to(self.opacity_effect, "opacity")
        self.fade_in_anim.set_range(1.0, 0.0)

    def display(self):
        """Display the notification."""
        self.message_display.setText(self.message)
        self.show()
        self.raise_()

    def close(self):
        """Close the notification."""
        super().close()
        self.deleteLater()

    def fade_in(self, duration: int):
        """Fade in the notification.

        Arguments:
            duration : int
                The desired duration of the animation

        Raises:
            TypeError: duration is not an integer
        """
        if type(duration) != int:
            raise TypeError("duration should be an integer")
        self.setGraphicsEffect(self.opacity_effect)
        self.fade_in_anim.setDuration(duration)
        self.is_fading_in = True
        self.fade_in_anim.finished.connect(self.on_fade_in_finished)
        self.display()
        self.fade_in_anim.start()

    def on_fade_in_finished(self):
        self.is_fading_in = False

    def fade_out(self, finishedCallback: Callable, duration: int):
        """Fade out the notification.

        Parameters
        ----------
        finishedCallback : callable
            The function to call after the animation has finished (to for instance
            clean up the notification)
        duration : int
            The desired duration of the animation

        Raises
        ------
        TypeError: the wrong datatype is specified for any of the parameters.

        """
        if not callable(finishedCallback):
            raise TypeError("finishedCallback should be a callable")
        if type(duration) != int:
            raise TypeError("duration should be an integer")

        self.setGraphicsEffect(self.opacity_effect)
        self.fade_out_anim.setDuration(duration)
        self.fade_out_anim.finished.connect(lambda: finishedCallback(self))
        self.is_being_removed = True
        self.fade_out_anim.start()

    def paintEvent(self, pe):
        """Redefinition of paintEvent, do not call directly.

        Makes class Notification available in style sheets. Interal Qt function.
        Should not be called directly.
        """
        o = widgets.StyleOption.based_on(self)
        p = gui.Painter(self)
        self.style().drawPrimitive(widgets.Style.PE_Widget, o, p, self)

    @property
    def message(self) -> str:
        """The currently set message to display."""
        return self._message

    @message.setter
    def message(self, value: str):
        """Set the message to display."""
        self._message = value

    @property
    def category(self) -> CategoryStr:
        """The currently set category of this notification."""
        return self._category

    @category.setter
    def category(self, value: CategoryStr):
        """Set the category of this notification.

        Arguments:
            value : {'primary','success','info','warning','danger'}
                The category specification

        Raises:
            ValueError: The category is other than one of the expected values.
        """
        allowed_values = ["primary", "success", "info", "warning", "danger"]
        if value not in allowed_values:
            raise ValueError(
                f"{value!r} is not a valid value. Should be one of {allowed_values}"
            )
        self._category = value

    def enterEvent(self, e):
        """When the notification is set to auto-hide, mouseover closes it."""
        if self.autohide:
            self.close_clicked.emit()


class NotificationArea(widgets.Widget):
    """Notification area to show notifications in.

    Will be projected on top of another QWidget which should be passed as an argument
    to this class.
    """

    def __init__(
        self,
        target_widget: QtWidgets.QWidget,
        use_global_css: bool = False,
        use_queue: bool = True,
        max_messages: int = 2,
    ):
        """Constructor.

        Arguments:
            target_widget : QtWidgets.QWidget
                The widget to project the notifications on
            use_global_css : bool (default: False)
                Flag which indicates whether global style sheets should be used
                (which have been set at app-level). If False, the default style sheets
                stored at DEFAULT_NOTIFICATION_STYLES will be loaded.
            use_queue : bool (default: True)
                Indicates whether a message queue should be implemented. This will only
                show *max_messages* at the same time and will put all other messages in a
                queue. Once a message disappears, the next one in the queue will be shown
                (up to max_messages at the same time)
            max_messages : int (default: 2)
                The number of messages to display at the same time.

        Raises:
            TypeError : target_widget is not an object that inherits QWidget
        """
        if not isinstance(target_widget, QtWidgets.QWidget):
            raise TypeError("target_widget is not a QWidget (or child of it")

        # Pop some variables from kwargs.
        self.use_queue = use_queue
        self.max_messages = max_messages
        super().__init__(parent=target_widget)

        if not use_global_css:
            self.setStyleSheet(DEFAULT_NOTIFICATION_STYLES)

        if self.use_queue:
            self.queue: Queue[Notification] = Queue()

        self.target_widget = target_widget
        self.set_margin(0)

        notification_area_layout = widgets.BoxLayout("vertical")
        self.setLayout(notification_area_layout)

        # Init effects to None
        self.entry_effect: FadeInValue = None
        self.entry_effect_duration = 250
        self.exit_effect: FadeOutValue = None
        self.exit_effect_duration = 500

        # Store original target classes resizeEvent to be called in our own
        # function
        self.target_resize_event = target_widget.resizeEvent
        # Overwrite resizeEvent function of target_widget to capture it ourself
        # (parent's resizeEvent will be called in our function too)
        self.target_widget.resizeEvent = self.resizeEvent  # type: ignore
        self.hide()

    def __delete_notification(self, notification=None):
        """Close and destroy the supplied notification."""
        notification.close()
        self.layout().removeWidget(notification)

        self.adjustSize()
        # Hide notification area if it doesn't contain any items
        if self.layout().count() == 0:
            self.hide()

        if self.use_queue:
            try:
                notification = self.queue.get(False)
                self._show_notification(notification)
            except Empty:
                pass

    # Public functions
    def set_entry_effect(self, effect: FadeInValue, duration: int = 250):
        """Set the effect with which the notifications are to appear.

        Arguments:
            effect : {'fade_in', None}
                The effect which should be used (for now only 'fade_in' is available)
                if None is passed for this argument, no effect will be used and the
                notifcations will just appear directly.
            duration : int (default: 250 ms)
                The duration of the effect in milliseconds

        Raises:
            TypeError: the object passed for duration is not an int
            ValueError: duration is less than 0, or effect has an invalid value
        """
        if effect not in ["fade_in", None]:
            raise ValueError("Invalid entry effect")
        if not isinstance(duration, int):
            raise TypeError("Duration should be an int")
        if duration < 0:
            raise ValueError("Duration should be larger than 0")

        self.entry_effect = effect
        self.entry_effect_duration = duration

    def set_exit_effect(self, effect: FadeOutValue, duration: int = 500):
        """Set the effect with which the notifications are to disappear.

        Arguments:
            effect : {'fade_out', None}
                the effect which should be used (for now only 'fade_out' is available)
                if None is passed for this argument, no effect will be used and the
                notifcations will just appear directly.
            duration : int (default: 1000 ms)
                The duration of the effect in milliseconds

        Raises:
            TypeError: the object passed for duration is not an int
            ValueError: duration is less than 0, or effect has an invalid value
        """
        if effect not in ["fade_out", None]:
            raise ValueError("Invalid exit effect")
        if not isinstance(duration, int):
            raise TypeError("Duration should be an int")
        if duration < 0:
            raise ValueError("Duration should be larger than 0")

        self.exit_effect = effect
        self.exit_effect_duration = duration

    @core.Slot(str, str, int, bool)
    @core.Slot(str, str, int, bool, str)
    def display(
        self,
        message: str,
        category: CategoryStr,
        timeout: int = 5000,
        autohide: bool = False,
        buttontext: Optional[str] = None,
    ):
        """Display a notification.

        If a queue is used, then the notification will only be shown directly
        if the number of notifications shown is smaller than max_messages.

        Arguments:
            message : str
                The message to display
            category : {'primary', 'success', 'info', 'warning', 'danger'}
                The type of notification that should be shown. Adheres to bootstrap
                standards which are primary, success, info, warning and danger
            timeout : int, optional
                The duration for which the notification should be shown. If None then
                the notification will be shown indefinitely
            buttontext : str, optional
                The text to display on the closing button. If not provided a cross
                will be shown.

        Raises:
            ValueError: the category is other than one of the expected values.
        """
        notification = Notification(
            message, category, timeout, autohide, buttontext, self
        )
        notification.close_clicked.connect(self.remove)

        # Queue if max amount of notifications is shown
        if self.use_queue and self.layout().count() >= self.max_messages:
            self.queue.put(notification)
        else:
            self._show_notification(notification)

    def _cursor_in_area(self):
        geom = self.geometry()
        top_left = self.mapToGlobal(geom.topLeft())
        bottom_right = self.mapToGlobal(geom.bottomRight())
        geom = core.Rect(top_left, bottom_right)
        geom.setTop(geom.top() - CURSOR_MARGIN_TOP)
        geom.setBottom(geom.bottom() + CURSOR_MARGIN_BOTTOM)
        geom.setLeft(geom.left() - CURSOR_MARGIN_LEFT)
        geom.setRight(geom.right() + CURSOR_MARGIN_RIGHT)
        cursor_pos = gui.Cursor.get_position()
        return geom.contains(cursor_pos)

    def _show_notification(self, notification: Notification):
        if self._cursor_in_area():
            core.Timer.singleShot(1000, lambda: self._show_notification(notification))
            return
        if not self.isVisible():
            self.show()
            self.raise_()
        self.layout().addWidget(notification)
        # Check for entry effects
        if self.entry_effect is not None:
            if self.entry_effect == "fade_in":
                notification.fade_in(self.entry_effect_duration)
        else:
            notification.display()

        self.adjustSize()
        if notification.timeout is not None and notification.timeout > 0:
            core.Timer.singleShot(notification.timeout, lambda: self.remove(notification))

    @core.Slot()
    def remove(self, notification: Optional[Notification] = None):
        """Removes a notification.

        Arguments:
            notification : Notification (default: None)
                The notification to remove. This function also serves as a PyQt slot
                for signals emitted from a Notification. In this case, the Notification
                object is retrieved by using self.sender()

        Raises:
            ValueError: notification is not None or a Notification

        """
        # This function also functions as a pyqt slot. In that case, no
        # notification argument is passed, but this is set as self.sender()
        if notification is None:
            notification = self.sender()
        if notification.is_being_removed or notification.is_fading_in:
            return
        notification.is_being_removed = True

        # Check if notification is still present (and has not manually been
        # closed before this function is called by a timeout)
        if notification not in self.layout():
            return

        # Implement animation here
        if self.exit_effect == "fade_out":
            notification.fade_out(self.__delete_notification, self.exit_effect_duration)
        else:
            self.__delete_notification(notification)

    # Internal Qt functions
    def resizeEvent(self, event):
        """Internal QT function (do not call directly)."""
        self.target_resize_event(event)
        newsize = event.size()
        self.setFixedWidth(newsize.width())
        self.adjustSize()

    def paintEvent(self, pe):
        """Redefinition of paintEvent.

        Makes class NotificationArea available in style sheets.
        Internal QT function (do not call directly).
        """
        o = widgets.StyleOption.based_on(self)
        p = gui.Painter(self)
        self.style().drawPrimitive(widgets.Style.PE_Widget, o, p, self)


if __name__ == "__main__":
    app = widgets.app()
    p = widgets.PushButton("test")
    noti = NotificationArea(p)
    p.clicked.connect(lambda: noti.display("test", "danger", timeout=2000))
    p.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtSvg


QtSvg.QGraphicsSvgItem.__bases__ = (widgets.GraphicsObject,)


class GraphicsSvgItem(QtSvg.QGraphicsSvgItem):
    pass



"""Webenginewidgets module.

contains QtWebEngineWidgets-based classes
"""

import logging

logger = logging.getLogger(__name__)

try:
    from .webenginehistoryitem import WebEngineHistoryItem
    from .webenginehistory import WebEngineHistory
    from .webenginesettings import WebEngineSettings
    from .webenginecontextmenudata import WebEngineContextMenuData
    from .webenginescript import WebEngineScript
    from .webenginescriptcollection import WebEngineScriptCollection
    from .webengineview import WebEngineView
    from .webenginepage import WebEnginePage
    from .webengineprofile import WebEngineProfile
    from .webenginedownloaditem import WebEngineDownloadItem

    __all__ = [
        "WebEngineHistoryItem",
        "WebEngineSettings",
        "WebEngineContextMenuData",
        "WebEngineHistory",
        "WebEngineView",
        "WebEnginePage",
        "WebEngineProfile",
        "WebEngineDownloadItem",
        "WebEngineScript",
        "WebEngineScriptCollection",
    ]
except ModuleNotFoundError:
    logger.warning("Tried to import WebEngine module but not installed.")



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QAbstractProxyModel.__bases__ = (core.AbstractItemModel,)


class AbstractProxyModel(QtCore.QAbstractProxyModel):
    pass



from __future__ import annotations

from typing import Iterator, List, Optional, Tuple

from prettyqt import gui
from prettyqt.qt import QtCore, QtGui


class CompositeValidator(gui.Validator):
    def __init__(
        self,
        validators: Optional[List[gui.Validator]] = None,
        parent: Optional[QtCore.QObject] = None,
    ):
        super().__init__(parent)
        self.validators = validators if validators is not None else []

    def __repr__(self):
        return f"{type(self).__name__}({self.validators})"

    def __getitem__(self, index: int) -> gui.Validator:
        return self.validators[index]

    def __setitem__(self, index: int, value: gui.Validator):
        self.validators[index] = value

    def __delitem__(self, index: int):
        del self.validators[index]

    def __contains__(self, index: int):
        return index in self.validators

    def __iter__(self) -> Iterator[gui.Validator]:
        return iter(self.validators)

    def __reduce__(self):
        return type(self), (self.validators,)

    def __len__(self):
        return len(self.validators)

    def __eq__(self, other: object):
        if not isinstance(other, type(self)):
            return False
        return self.validators == other.validators

    def validate(  # type: ignore
        self, text: str, pos: int = 0
    ) -> Tuple[QtGui.QValidator.State, str, int]:
        vals = [v.validate(text, pos)[0] for v in self.validators]  # type: ignore
        return min(vals), text, pos


if __name__ == "__main__":
    from prettyqt import custom_validators, widgets

    val1 = custom_validators.NotEmptyValidator()
    val2 = custom_validators.PathValidator()
    val = CompositeValidator([val1, val2])
    app = widgets.app()
    widget = widgets.LineEdit("This is a test")
    widget.setValidator(val)
    widget.show()
    app.main_loop()



from __future__ import annotations

from typing import Any, Iterable, Literal, Mapping, Optional, Tuple, Union

from prettyqt import core, gui, iconprovider, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, bidict, types


INSERT_POLICY = bidict(
    no_insert=QtWidgets.QComboBox.NoInsert,
    top=QtWidgets.QComboBox.InsertAtTop,
    current=QtWidgets.QComboBox.InsertAtCurrent,
    bottom=QtWidgets.QComboBox.InsertAtBottom,
    after_current=QtWidgets.QComboBox.InsertAfterCurrent,
    before_current=QtWidgets.QComboBox.InsertBeforeCurrent,
    alphabetically=QtWidgets.QComboBox.InsertAlphabetically,
)

InsertPolicyStr = Literal[
    "no_insert",
    "top",
    "current",
    "bottom",
    "after_current",
    "before_current",
    "alphabetically",
]

SIZE_ADJUST_POLICY = bidict(
    content=QtWidgets.QComboBox.AdjustToContents,
    first_show=QtWidgets.QComboBox.AdjustToContentsOnFirstShow,
    # min_length=QtWidgets.QComboBox.AdjustToMinimumContentsLength,
    min_length_with_icon=QtWidgets.QComboBox.AdjustToMinimumContentsLengthWithIcon,
)

SizeAdjustPolicyStr = Literal[
    "content",
    "first_show",
    # "min_length",
    "min_length_with_icon",
]


class NoData:
    pass


QtWidgets.QComboBox.__bases__ = (widgets.Widget,)


class ComboBox(QtWidgets.QComboBox):

    value_changed = core.Signal(object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.currentIndexChanged.connect(self.index_changed)

    def serialize_fields(self):
        items = [
            (self.itemText(i), self.itemData(i), self.item_icon(i))
            for i in range(self.count())
        ]
        dct = dict(
            index=self.currentIndex(),
            editable=self.isEditable(),
            max_count=self.maxCount(),
            has_frame=self.hasFrame(),
            icon_size=self.iconSize(),
            insert_policy=self.insertPolicy(),
            model_column=self.modelColumn(),
            max_visible_items=self.maxVisibleItems(),
            size_adjust_policy=self.get_size_adjust_policy(),
            duplicates_enabled=self.duplicatesEnabled(),
            min_contents_length=self.minimumContentsLength(),
            items=items,
        )
        if core.VersionNumber.get_qt_version() >= (5, 15, 0):
            dct["placeholder_text"] = self.placeholderText()
        return dct

    def __setstate__(self, state):
        super().__setstate__(state)
        for label, data, icon in state["items"]:
            self.add(label, data, icon=icon)
        self.setCurrentIndex(state["index"])
        self.setEditable(state["editable"])
        self.setMaxCount(state["max_count"])
        self.setMaxVisibleItems(state["max_visible_items"])
        self.setMinimumContentsLength(state["min_contents_length"])
        self.setDuplicatesEnabled(state["duplicates_enabled"])
        self.setFrame(state["has_frame"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __len__(self) -> int:
        return self.count()

    def index_changed(self, index: int):
        data = self.itemData(index)
        self.value_changed.emit(data)

    def add_items(self, items: Union[Iterable, Mapping], default=NoData):
        if isinstance(items, Mapping):
            for k, v in items.items():
                self.addItem(v, userData=k)
        else:
            for i in items:
                if isinstance(i, (tuple, list)):
                    self.add(*i)
                else:
                    self.addItem(i, i)
        if default is not NoData:
            self.set_value(default)

    def add(self, label: str, data=NoData, icon: types.IconType = None):
        if data is NoData:
            data = label
        if icon is not None:
            icon = iconprovider.get_icon(icon)
            self.addItem(gui.Icon(icon), label, userData=data)
        else:
            self.addItem(label, userData=data)

    def item_icon(self, index: int) -> Optional[gui.Icon]:
        icon = self.itemIcon(index)
        if icon.isNull():
            return None
        return gui.Icon(icon)

    def set_insert_policy(self, policy: InsertPolicyStr):
        """Set insert policy.

        Args:
            policy: insert policy to use

        Raises:
            InvalidParamError: invalid insert policy
        """
        if policy not in INSERT_POLICY:
            raise InvalidParamError(policy, INSERT_POLICY)
        self.setInsertPolicy(INSERT_POLICY[policy])

    def get_insert_policy(self) -> InsertPolicyStr:
        """Return insert policy.

        Returns:
            insert policy
        """
        return INSERT_POLICY.inverse[self.insertPolicy()]

    def set_size_adjust_policy(self, policy: SizeAdjustPolicyStr):
        """Set size adjust policy.

        Args:
            policy: size adjust policy to use

        Raises:
            InvalidParamError: invalid size adjust policy
        """
        if policy not in SIZE_ADJUST_POLICY:
            raise InvalidParamError(policy, SIZE_ADJUST_POLICY)
        self.setSizeAdjustPolicy(SIZE_ADJUST_POLICY[policy])

    def get_size_adjust_policy(self) -> SizeAdjustPolicyStr:
        """Return size adjust policy.

        Returns:
            size adjust policy
        """
        return SIZE_ADJUST_POLICY.inverse[self.sizeAdjustPolicy()]

    def set_icon_size(self, size: Union[int, Tuple[int, int], QtCore.QSize]):
        """Set size of the icons."""
        if isinstance(size, int):
            size = core.Size(size, size)
        elif isinstance(size, tuple):
            size = core.Size(*size)
        self.setIconSize(size)

    def get_icon_size(self) -> core.Size:
        return core.Size(self.iconSize())

    def set_min_char_length(self, chars: int):
        self.setMinimumContentsLength(chars)

    def get_value(self) -> Any:
        # if all(self.itemData(i) is None for i in range(self.count())):
        #     return self.currentText()
        # else:
        #     return self.currentData()
        return self.currentData()

    def set_value(self, value: Any):
        self.set_data(value)

    def set_text(self, text: str):
        self.setCurrentText(text)

    def set_data(self, data: Any):
        idx = self.findData(data)
        if idx == -1:
            raise ValueError("invalid data")
        self.setCurrentIndex(idx)

    def text(self) -> str:
        return self.currentText()


if __name__ == "__main__":
    app = widgets.app()
    widget = ComboBox()
    widget.value_changed.connect(print)
    w = ComboBox()
    widget.add("test", data="aa", icon="mdi.timer")
    widget.add("test2", data="aa2", icon="mdi.timer")
    widget.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import gui, iconprovider
from prettyqt.qt import QtWidgets
from prettyqt.utils import bidict, types


ACTIVATION_REASONS = bidict(
    unknown=QtWidgets.QSystemTrayIcon.Unknown,
    context=QtWidgets.QSystemTrayIcon.Context,
    double_click=QtWidgets.QSystemTrayIcon.DoubleClick,
    trigger=QtWidgets.QSystemTrayIcon.Trigger,
    middle_click=QtWidgets.QSystemTrayIcon.MiddleClick,
)

MESSAGE_ICONS = bidict(
    none=QtWidgets.QSystemTrayIcon.NoIcon,
    information=QtWidgets.QSystemTrayIcon.Information,
    warning=QtWidgets.QSystemTrayIcon.Warning,
    critical=QtWidgets.QSystemTrayIcon.Critical,
)


class SystemTrayIcon(QtWidgets.QSystemTrayIcon):
    def set_icon(self, icon: types.IconType):
        """Set the system tray icon.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon)
        self.setIcon(icon)

    def show_message(
        self,
        title: str,
        message: str = "",
        icon: types.IconType = None,
        timeout: int = 10,
    ):
        if icon is None:
            ico = gui.Icon()
        if icon in MESSAGE_ICONS:
            ico = MESSAGE_ICONS[icon]
        else:
            ico = iconprovider.get_icon(icon)
        self.showMessage(title, message, ico, timeout * 1000)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    icon = SystemTrayIcon()
    icon.set_icon("mdi.folder")
    print(icon.isSystemTrayAvailable())
    icon.show_message("a", "b", "information")
    app.main_loop()



from __future__ import annotations

from typing import List

from prettyqt import positioning
from prettyqt.qt import QtPositioning


QtPositioning.QGeoPath.__bases__ = (positioning.GeoShape,)


class GeoPath(QtPositioning.QGeoPath):
    def __len__(self):
        return self.size()

    def __getitem__(self, index: int) -> positioning.GeoCoordinate:
        return positioning.GeoCoordinate(self.coordinateAt(index))

    def __setitem__(self, index: int, value: QtPositioning.QGeoCoordinate):
        self.replaceCoordinate(index, value)

    def __delitem__(self, index: int):
        self.removeCoordinate(index)

    def __add__(self, other: QtPositioning.QGeoCoordinate):
        self.addCoordinate(other)
        return self

    def __repr__(self):
        # p = ", ".join([f"{p!r}" for p in self.get_path()])
        return f"{type(self).__name__}(<{len(self)} points>)"

    def get_path(self) -> List[positioning.GeoCoordinate]:
        return [positioning.GeoCoordinate(p) for p in self.path()]


if __name__ == "__main__":
    path = GeoPath()
    print(str(path))



from __future__ import annotations

import pathlib

from prettyqt import gui


class PathValidator(gui.Validator):
    def __eq__(self, other: object):
        return isinstance(other, PathValidator)

    def validate(self, text: str, pos: int = 0):
        if pathlib.Path(text).exists():
            return self.Acceptable, text, pos
        return self.Intermediate, text, pos


if __name__ == "__main__":
    from prettyqt import widgets

    val = PathValidator()
    app = widgets.app()
    widget = widgets.LineEdit("This is a test")
    widget.setValidator(val)
    widget.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWidgets


QtWidgets.QMouseEventTransition.__bases__ = (core.EventTransition,)


class MouseEventTransition(QtWidgets.QMouseEventTransition):
    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtMultimedia


class MediaContent(QtMultimedia.QMediaContent):
    def get_url(self):
        return core.Url(self.canonicalUrl())



from __future__ import annotations

import functools
import operator
import sys
from typing import Any, Dict, List, Mapping


def string_to_num_array(array: str) -> List[float]:
    floats = [float(i) for i in array.split(",")]
    return [int(i) if i.is_integer() else i for i in floats]


def merge_flags(flags, mapping: Mapping):
    return functools.reduce(operator.ior, [mapping[t] for t in flags])


def format_kwargs(kwargs: Dict[str, Any]) -> str:
    kwarg_list = [f"{k}={repr(v)}" for k, v in kwargs.items()]
    return ", ".join(kwarg_list)


def cut_off_str(obj, max_len: int) -> str:
    """Create a string representation of an object, no longer than max_len characters.

    Uses repr(obj) to create the string representation. If this is longer than max_len -3
    characters, the last three will be replaced with elipsis.
    """
    s = repr(obj)
    if len(s) > max_len - 3:
        return s[: max_len - 3] + "..."
    return s


def is_dark_mode() -> bool:
    if sys.platform.startswith("win"):
        from prettyqt import core

        p = (
            "HKEY_CURRENT_USER\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\"
            "Themes\\Personalize"
        )
        s = core.Settings(p, core.Settings.NativeFormat)
        return s.value("AppsUseLightTheme") == 0
    elif sys.platform == "darwin":
        import darkdetect

        return darkdetect.isDark()
    else:
        return False



from __future__ import annotations

from prettyqt import gui
from prettyqt.qt import QtGui


class FontInfo(QtGui.QFontInfo):
    def get_style_hint(self) -> gui.font.StyleHintStr:
        return gui.font.STYLE_HINTS.inverse[self.styleHint()]



from __future__ import annotations

from prettyqt import location
from prettyqt.qt import QtLocation


# not available in PySide2

QtLocation.QGeoRouteLeg.__bases__ = (location.GeoRoute,)  # type: ignore


class GeoRouteLeg(QtLocation.QGeoRouteLeg):  # type: ignore
    pass



from __future__ import annotations

from typing import Literal

from prettyqt import widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import bidict


SWIPE_DIRECTION = bidict(
    none=QtWidgets.QSwipeGesture.NoDirection,
    left=QtWidgets.QSwipeGesture.Left,
    right=QtWidgets.QSwipeGesture.Right,
    up=QtWidgets.QSwipeGesture.Up,
    down=QtWidgets.QSwipeGesture.Down,
)

HorizontalDirectionStr = Literal["none", "left", "right"]
VerticalDirectionStr = Literal["none", "up", "down"]

QtWidgets.QSwipeGesture.__bases__ = (widgets.Gesture,)


class SwipeGesture(QtWidgets.QSwipeGesture):
    def get_horizontal_direction(self) -> HorizontalDirectionStr:
        """Return horizontal direction of the gesture.

        Returns:
            horizontal direction
        """
        return SWIPE_DIRECTION.inverse[self.horizontalDirection()]

    def get_vertical_direction(self) -> VerticalDirectionStr:
        """Return vertical direction of the gesture.

        Returns:
            vertical direction
        """
        return SWIPE_DIRECTION.inverse[self.verticalDirection()]


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    gesture = SwipeGesture()



from __future__ import annotations

from typing import Tuple, Union

from prettyqt import core
from prettyqt.qt import QtWidgets


class BounceAnimation(core.SequentialAnimationGroup):
    def __init__(
        self, duration: int = 1000, easing: core.easingcurve.TypeStr = "in_out_sine"
    ):
        super().__init__()
        self.anim1 = core.PropertyAnimation()
        self.anim2 = core.PropertyAnimation()
        self.set_easing(easing)
        self.set_start_value((0, 0))
        # self.set_end_value(core.Point(0, 100))
        self.addAnimation(self.anim1)
        self.addAnimation(self.anim2)
        self.set_duration(duration)

    def set_start_value(self, point: Union[core.Point, Tuple[int, int]]):
        if isinstance(point, tuple):
            point = core.Point(*point)
        self.anim1.setStartValue(point)
        self.anim2.setEndValue(point)

    def set_end_value(self, point: Union[core.Point, Tuple[int, int]]):
        if isinstance(point, tuple):
            point = core.Point(*point)
        self.anim2.setStartValue(point)
        self.anim1.setEndValue(point)

    def set_duration(self, duration: int):
        self.anim1.setDuration(duration // 2)
        self.anim2.setDuration(duration // 2)

    def set_easing(self, easing: core.easingcurve.TypeStr):
        self.anim1.set_easing(easing)
        self.anim2.set_easing(easing)

    def apply_to(self, obj: QtWidgets.QWidget):
        self.anim1.apply_to(obj, "pos")
        self.anim2.apply_to(obj, "pos")


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    val = BounceAnimation()



from __future__ import annotations

from prettyqt import core, location
from prettyqt.qt import QtLocation


QtLocation.QPlaceReview.__bases__ = (location.PlaceContent,)


class PlaceReview(QtLocation.QPlaceReview):
    def __str__(self):
        return f"{self.title()}: {self.text()}"

    def get_datetime(self) -> core.DateTime:
        return core.DateTime(self.dateTime())


if __name__ == "__main__":
    review = PlaceReview()



from __future__ import annotations

from prettyqt import widgets


class MarkdownWindow(widgets.MainWindow):
    def __init__(self):
        super().__init__()
        self.resize(500, 500)
        self.text_browser = widgets.TextBrowser()
        self.setCentralWidget(self.text_browser)
        # self.text_browser.loadFinished.connect(self._load_finished)
        self.create_menu()

    # def _load_finished(self):
    # frame = self.text_browser.page()
    # self.text_browser.page().setViewportSize(frame.contentsSize())
    # self.resize(frame.contentsSize())
    # html_data = frame.toHtml()

    def create_menu(self):
        act_exit = widgets.Action(
            text="&Exit",
            icon="mdi.exit-to-app",
            parent=self,
            shortcut="Ctrl+Q",
            statustip="Exit application",
            callback=self.close,
        )
        act_open = widgets.Action(
            text="&Open",
            icon="mdi.open-in-app",
            parent=self,
            shortcut="Ctrl+O",
            statustip="Open Markdown file",
            callback=self.open_new_file,
        )
        menubar = self.menuBar()
        menu_file = menubar.addMenu("&File")
        menu_file.addAction(act_open)
        menu_file.addAction(act_exit)

    # def dragEnterEvent(self, event):
    #     u = event.mimeData().urls()
    #     for url in u:
    #         file_path = os.path.abspath(url.toLocalFile())

    #         ext = file_path.split(".")[-1]
    #         if ext in ["txt", "md", "markdown"]:
    #             event.accept()
    #         else:
    #             event.ignore()

    # def dropEvent(self, event):
    #     event.accept()
    #     self.show_markdown(self.filePath)

    def open_new_file(self):
        try:
            ext = {"All Text Files": [".md", ".markdown", ".txt"]}
            dlg = widgets.FileDialog(mode="open", extension_filter=ext)
            if (fname := dlg.open_file()) is not None:
                self.text_browser.set_markdown_file(fname[0])
        except UnicodeDecodeError:
            self.statusBar().showMessage("Please select only text files")
        except OSError:
            self.statusBar().showMessage("File open canceled!")


if __name__ == "__main__":
    app = widgets.app()
    reader = MarkdownWindow()
    # reader.text_browser.load(QtCore.QUrl("blank"))
    reader.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QStyleOptionSizeGrip.__bases__ = (widgets.StyleOptionComplex,)


class StyleOptionSizeGrip(QtWidgets.QStyleOptionSizeGrip):
    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QConcatenateTablesProxyModel.__bases__ = (core.AbstractItemModel,)


class ConcatenateTablesProxyModel(QtCore.QConcatenateTablesProxyModel):

    pass



from __future__ import annotations

from typing import Optional, Tuple, Union

from prettyqt import core, gui, iconprovider, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets
from prettyqt.utils import InvalidParamError, types


QtWidgets.QAbstractButton.__bases__ = (widgets.Widget,)


class AbstractButton(QtWidgets.QAbstractButton):
    def serialize_fields(self):
        return dict(
            text=self.text(),
            icon=self.get_icon(),
            checkable=self.isCheckable(),
            checked=self.isChecked(),
            auto_exclusive=self.autoExclusive(),
            auto_repeat=self.autoRepeat(),
            auto_repeat_delay=self.autoRepeatDelay(),
            auto_repeat_interval=self.autoRepeatInterval(),
            is_down=self.isDown(),
            icon_size=self.get_icon_size(),
            shortcut=self.get_shortcut(),
        )

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setText(state["text"])
        self.set_icon(state["icon"])
        self.setChecked(state.get("checked", False))
        self.setCheckable(state["checkable"])
        self.setAutoExclusive(state["auto_exclusive"])
        self.setAutoRepeat(state["auto_repeat"])
        self.setAutoRepeatDelay(state["auto_repeat_delay"])
        self.setAutoRepeatInterval(state["auto_repeat_interval"])
        self.setDown(state["is_down"])
        self.set_icon_size(state["icon_size"])
        self.setShortcut(state["shortcut"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __bool__(self):
        return self.isChecked()

    def set_icon(self, icon: types.IconType):
        """Set the icon for the button.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon)
        self.setIcon(icon)

    def get_icon(self) -> Optional[gui.Icon]:
        icon = self.icon()
        if icon.isNull():
            return None
        return gui.Icon(icon)

    def set_style_icon(self, icon: widgets.style.StandardPixmapStr, size: int = 15):
        """Set theme icon for the button.

        Args:
            icon: icon to use
            size: icon size
        """
        if icon not in widgets.style.STANDARD_PIXMAP:
            raise InvalidParamError(icon, widgets.style.STANDARD_PIXMAP)
        qicon = self.style().standardIcon(widgets.style.STANDARD_PIXMAP[icon], None, self)
        self.set_icon(qicon)
        self.setIconSize(core.Size(size, size))

    def set_shortcut(self, shortcut: Union[None, QtGui.QKeySequence, str]):
        if shortcut is None:
            shortcut = ""
        if isinstance(shortcut, str):
            shortcut = gui.KeySequence(shortcut, gui.KeySequence.PortableText)
        self.setShortcut(shortcut)

    def get_shortcut(self) -> gui.KeySequence:
        return gui.KeySequence(self.shortcut().toString(), gui.KeySequence.PortableText)

    def set_text(self, text: str):
        self.setText(text)

    def set_icon_size(self, size: Union[int, Tuple[int, int], QtCore.QSize]):
        """Set size of the icon."""
        if isinstance(size, int):
            size = core.Size(size, size)
        elif isinstance(size, tuple):
            size = core.Size(*size)
        self.setIconSize(size)

    def get_icon_size(self) -> core.Size:
        return core.Size(self.iconSize())

    def get_value(self) -> bool:
        return self.isChecked()

    def set_value(self, value: bool):
        self.setChecked(value)

    @property
    def is_on(self) -> bool:
        return self.isChecked()

    @is_on.setter
    def is_on(self, state: bool):
        self.setChecked(state)


if __name__ == "__main__":
    app = widgets.app()
    widget = AbstractButton()
    widget.show()
    app.main_loop()



from __future__ import annotations

from prettyqt.qt import QtCore


class JsonValue(QtCore.QJsonValue):
    def __str__(self):
        return str(self.toVariant())

    def __repr__(self):
        return f"{type(self).__name__}({self.toVariant()!r})"


if __name__ == "__main__":
    from prettyqt import core

    doc = core.JsonDocument.from_variant(dict(a="b"))
    print(doc.toVariant())
    print(str(doc))
    new = core.JsonDocument()
    new.setObject(doc.object())
    print("test")
    val = doc["a"]
    print(type(val), val)



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QGraphicsItemGroup.__bases__ = (widgets.GraphicsItem,)


class GraphicsItemGroup(QtWidgets.QGraphicsItemGroup):
    pass



from __future__ import annotations

from typing import Optional, Union

from prettyqt import core, widgets
from prettyqt.qt import QtCore, QtWidgets


class ExpandableLine(widgets.Widget):
    def __init__(
        self,
        title: str = "",
        animation_duration: int = 300,
        parent: Optional[QtWidgets.QWidget] = None,
    ):
        super().__init__(parent=parent)

        self._animation_duration = animation_duration

        self.expand_btn = widgets.ToolButton()
        self.expand_btn.set_text(title)
        with self.expand_btn.edit_stylesheet() as ss:
            ss.QToolButton.border.setValue(None)
        self.expand_btn.set_style("text_beside_icon")
        self.expand_btn.set_arrow_type("right")
        self.expand_btn.setCheckable(True)
        self.expand_btn.setChecked(False)

        header_line = widgets.Frame()
        header_line.set_frame_shape("h_line")
        header_line.set_frame_shadow("sunken")
        header_line.set_size_policy("expanding", "maximum")

        self.content_area = widgets.ScrollArea()
        with self.expand_btn.edit_stylesheet() as ss:
            ss.QAbstractScrollArea.border.setValue(None)
        self.content_area.set_size_policy("expanding", "fixed")
        self.content_area.setMaximumHeight(1)
        # self.content_area.setMinimumHeight(0)

        self.toggle_anim = core.ParallelAnimationGroup()
        self.toggle_anim.add_property_animation(self, "minimumHeight")
        self.toggle_anim.add_property_animation(self, "maximumHeight")
        self.toggle_anim.add_property_animation(self.content_area, "maximumHeight")
        base_layout = widgets.GridLayout()
        base_layout.setVerticalSpacing(0)
        base_layout.set_margin(0)
        base_layout.addWidget(self.expand_btn, 0, 0, 1, 1, QtCore.Qt.AlignLeft)
        base_layout[0, 2] = header_line
        base_layout[1, 0:2] = self.content_area
        self.setLayout(base_layout)
        # self.toggle_anim.setStartValue(0)
        # self.toggle_anim.setEndValue(300)

        def expand_view(checked: bool):
            self.expand_btn.set_arrow_type("down" if checked else "right")
            self.toggle_anim.set_direction("forward" if checked else "backward")
            self.toggle_anim.start()

        # === SIGNALS === #
        self.expand_btn.toggled.connect(expand_view)
        self.toggle_anim.set_duration(0)
        self.toggle_anim.set_duration(self._animation_duration)

    def set_layout(
        self,
        layout: Union[str, QtWidgets.QLayout, None],
        margin: Optional[int] = None,
        spacing: Optional[int] = None,
    ) -> None:
        self.content_area.destroy()
        self.content_area.set_layout(layout, margin=margin, spacing=spacing)
        collapsed_height = self.sizeHint().height() - self.content_area.maximumHeight()
        content_height = self.content_area.box.sizeHint().height() + 300
        for expand_anim in self.toggle_anim[:-1]:
            # expand_anim.setDuration(self._animation_duration)
            expand_anim.set_range(collapsed_height, collapsed_height + content_height)
        content_anim = self.toggle_anim[-1]
        content_anim.set_range(1, content_height)


if __name__ == "__main__":
    app = widgets.app()
    layout = widgets.BoxLayout("vertical")
    layout.addWidget(widgets.TextBrowser())
    widget = ExpandableLine("Test")
    widget.set_layout(layout)
    widget.show()
    app.main_loop()



from __future__ import annotations

from typing import Optional, Union

from prettyqt import constants
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError


class LayoutItem(QtWidgets.QLayoutItem):
    # def __bool__(self):
    #     return not self.isEmpty()

    def set_alignment(self, alignment: constants.AlignmentStr):
        """Set the alignment of the layout.

        Args:
            alignment: alignment for the layout

        Raises:
            InvalidParamError: alignment does not exist
        """
        if alignment not in constants.ALIGNMENTS:
            raise InvalidParamError(alignment, constants.ALIGNMENTS)
        self.setAlignment(constants.ALIGNMENTS[alignment])

    def get_alignment(self) -> constants.AlignmentStr:
        """Return current alignment.

        Returns:
            alignment
        """
        return constants.ALIGNMENTS.inverse[self.alignment()]

    def get_item(
        self,
    ) -> Optional[Union[QtWidgets.QWidget, QtWidgets.QLayout, QtWidgets.QSpacerItem]]:
        if content := self.widget():
            return content
        if content := self.layout():
            return content
        if content := self.spacerItem():
            return content
        return None



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QTransposeProxyModel.__bases__ = (core.AbstractProxyModel,)


class TransposeProxyModel(QtCore.QTransposeProxyModel):
    pass



from __future__ import annotations

from typing import Literal, Union

from prettyqt import network
from prettyqt.qt import QtNetwork
from prettyqt.utils import bidict


CONVERSION_MODE = bidict(
    strict=QtNetwork.QHostAddress.StrictConversion,
    v4_mapped_to_ipv4=QtNetwork.QHostAddress.ConvertV4MappedToIPv4,
    v4_compat_to_ipv4=QtNetwork.QHostAddress.ConvertV4CompatToIPv4,
    localhost=QtNetwork.QHostAddress.ConvertLocalHost,
    unspecified_address=QtNetwork.QHostAddress.ConvertUnspecifiedAddress,
    tolerant=QtNetwork.QHostAddress.TolerantConversion,
)

SPECIAL_ADDRESS = bidict(
    null=int(QtNetwork.QHostAddress.Null),
    localhost=int(QtNetwork.QHostAddress.LocalHost),
    localhost_ipv6=int(QtNetwork.QHostAddress.LocalHostIPv6),
    broadcast=int(QtNetwork.QHostAddress.Broadcast),
    any_ipv4=int(QtNetwork.QHostAddress.AnyIPv4),
    any_ipv6=int(QtNetwork.QHostAddress.AnyIPv6),
    any=int(QtNetwork.QHostAddress.Any),
)

SpecialAddressStr = Literal[
    "null", "localhost", "localhost_ipv6", "broadcast", "any_ipv4", "any_ipv6", "any"
]

NetworkLayerProtocolStr = Literal["ipv4", "ipv6", "any_ip", "unknown"]


class HostAddress(QtNetwork.QHostAddress):
    def __repr__(self):
        return f"{type(self).__name__}({self.toString()!r})"

    def __str__(self):
        return self.toString()

    def __bool__(self):
        return not self.isNull()

    def get_protocol(self) -> NetworkLayerProtocolStr:
        return network.abstractsocket.NETWORK_LAYER_PROTOCOL.inverse[self.protocol()]

    def set_address(self, address: Union[int, str]):
        if address in SPECIAL_ADDRESS:
            address = SPECIAL_ADDRESS[address]
        ret = self.setAddress(address)
        if ret is False:
            raise ValueError("invalid address")



from __future__ import annotations

import logging
import pathlib
from typing import Union

from pygments.formatters.html import HtmlFormatter
from pygments.lexer import Error, RegexLexer, Text, _TokenType
from pygments.lexers import get_lexer_by_name, load_lexer_from_file
from pygments.style import Style
from pygments.styles import get_style_by_name

from prettyqt import gui
from prettyqt.qt import QtGui


logger = logging.getLogger(__name__)


def qstring_length(text: str) -> int:
    """Tries to compute what the length of an utf16-encoded QString would be."""
    utf16_text = text.encode("utf16")
    length = len(utf16_text) // 2
    # Remove Byte order mark.
    # TODO: All unicode Non-characters should be removed
    if utf16_text[:2] in [b"\xff\xfe", b"\xff\xff", b"\xfe\xff"]:
        length -= 1
    return length


def get_tokens_unprocessed(self, text: str, stack=("root",)):
    """Split ``text`` into (tokentype, text) pairs.

    Monkeypatched to store the final stack on the object itself.

    The `text` parameter this gets passed is only the current line, so to
    highlight things like multiline strings correctly, we need to retrieve
    the state from the previous line (this is done in PygmentsHighlighter,
    below), and use it to continue processing the current line.
    """
    pos = 0
    tokendefs = self._tokens
    if hasattr(self, "_saved_state_stack"):
        statestack = list(self._saved_state_stack)
    else:
        statestack = list(stack)
    statetokens = tokendefs[statestack[-1]]
    while 1:
        for rexmatch, action, new_state in statetokens:
            m = rexmatch(text, pos)
            if not m:
                continue
            if action is not None:
                if type(action) is _TokenType:
                    yield pos, action, m.group()
                else:
                    yield from action(self, m)
            pos = m.end()
            if new_state is None:
                break
            # state transition
            if isinstance(new_state, tuple):
                for state in new_state:
                    if state == "#pop":
                        statestack.pop()
                    elif state == "#push":
                        statestack.append(statestack[-1])
                    else:
                        statestack.append(state)
            elif isinstance(new_state, int):
                # pop
                del statestack[new_state:]
            elif new_state == "#push":
                statestack.append(statestack[-1])
            else:
                assert False, "wrong state def: %r" % new_state
            statetokens = tokendefs[statestack[-1]]
            break
        else:
            try:
                if text[pos] == "\n":
                    # at EOL, reset state to "root"
                    pos += 1
                    statestack = ["root"]
                    statetokens = tokendefs["root"]
                    yield pos, Text, "\n"
                    continue
                yield pos, Error, text[pos]
                pos += 1
            except IndexError:
                break
    self._saved_state_stack = list(statestack)


# Monkeypatch!
RegexLexer.get_tokens_unprocessed = get_tokens_unprocessed


class PygmentsHighlighter(gui.SyntaxHighlighter):
    """Syntax highlighter that uses Pygments for parsing."""

    # ---------------------------------------------------------------------------
    #  "QSyntaxHighlighter" interface
    # ---------------------------------------------------------------------------

    def __init__(self, parent: QtGui.QTextDocument, lexer: str):
        super().__init__(parent)
        self._document = self.document()
        self._formatter = HtmlFormatter(nowrap=True)
        self.set_style("default")
        if lexer == "regex":
            path = pathlib.Path(__file__).parent
            path = path / "pygments" / "regularexpressionlexer.py"
            self._lexer = load_lexer_from_file(str(path))
        else:
            self._lexer = get_lexer_by_name(lexer)

    def highlightBlock(self, string):
        """Highlight a block of text."""
        prev_data = self.currentBlock().previous().userData()
        if prev_data is not None:
            self._lexer._saved_state_stack = prev_data.syntax_stack
        elif hasattr(self._lexer, "_saved_state_stack"):
            del self._lexer._saved_state_stack

        # Lex the text using Pygments
        index = 0
        for token, text in self._lexer.get_tokens(string):
            length = qstring_length(text)
            self.setFormat(index, length, self._get_format(token))
            index += length

        if hasattr(self._lexer, "_saved_state_stack"):
            data = gui.TextBlockUserData(syntax_stack=self._lexer._saved_state_stack)
            self.currentBlock().setUserData(data)
            # Clean up for the next go-round.
            del self._lexer._saved_state_stack

    # ---------------------------------------------------------------------------
    # "PygmentsHighlighter" interface
    # ---------------------------------------------------------------------------

    def set_style(self, style: Union[str, Style]):
        """Sets the style to the specified Pygments style."""
        if isinstance(style, str):
            style = get_style_by_name(style)
        self._style = style
        self._clear_caches()

    def set_style_sheet(self, stylesheet: str):
        """Sets a CSS stylesheet.

        The classes in the stylesheet should correspond to those generated by:

            pygmentize -S <style> -f html

        Note that "set_style" and "set_style_sheet" completely override each
        other, i.e. they cannot be used in conjunction.
        """
        self._document.setDefaultStyleSheet(stylesheet)
        self._style = None
        self._clear_caches()

    # ---------------------------------------------------------------------------
    # Protected interface
    # ---------------------------------------------------------------------------

    def _clear_caches(self):
        """Clear caches for brushes and formats."""
        self._brushes = {}
        self._formats = {}

    def _get_format(self, token: str) -> QtGui.QTextCharFormat:
        """Returns a QTextCharFormat for token or None."""
        if token in self._formats:
            return self._formats[token]

        if self._style is None:
            result = self._get_format_from_document(token, self._document)
        else:
            result = self._get_format_from_style(token, self._style)

        self._formats[token] = result
        return result

    def _get_format_from_document(
        self, token: str, document: QtGui.QTextDocument
    ) -> QtGui.QTextCharFormat:
        """Return a QTextCharFormat for token from document."""
        code, html = next(self._formatter._format_lines([(token, "dummy")]))
        self._document.setHtml(html)
        return gui.TextCursor(self._document).charFormat()

    def _get_format_from_style(self, token: str, style: Style) -> gui.TextCharFormat:
        """Return a QTextCharFormat for token by reading a Pygments style."""
        result = gui.TextCharFormat()
        try:
            token_style = style.style_for_token(token)
        except KeyError:
            return result
        for key, value in token_style.items():
            if value:
                if key == "color":
                    result.set_foreground_color(self._get_brush(value))
                elif key == "bgcolor":
                    result.set_background_color(self._get_brush(value))
                elif key == "bold":
                    result.set_font_weight("bold")
                elif key == "italic":
                    result.setFontItalic(True)
                elif key == "underline":
                    result.set_underline_style("single")
                elif key == "sans":
                    result.set_font_style_hint("sans_serif")
                elif key == "roman":
                    result.set_font_style_hint("serif")
                elif key == "mono":
                    result.set_font_style_hint("typewriter")
        return result

    def _get_brush(self, color: str) -> gui.Brush:
        """Return a brush for the color."""
        result = self._brushes.get(color)
        if result is None:
            qcolor = gui.Color(f"#{color[:6]}")
            result = gui.Brush(qcolor)
            self._brushes[color] = result
        return result


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    editor = widgets.PlainTextEdit()
    highlighter = PygmentsHighlighter(editor.document(), lexer="regex")
    editor.show()
    app.main_loop()



from __future__ import annotations

from typing import Literal, Optional

from prettyqt import constants, core, gui, widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError, bidict


ECHO_MODE = bidict(
    normal=QtWidgets.QLineEdit.Normal,
    no_echo=QtWidgets.QLineEdit.NoEcho,
    password=QtWidgets.QLineEdit.Password,
    echo_on_edit=QtWidgets.QLineEdit.PasswordEchoOnEdit,
)

EchoModeStr = Literal["normal", "no_echo", "password", "echo_on_edit"]

ACTION_POSITION = bidict(
    leading=QtWidgets.QLineEdit.LeadingPosition,
    trailing=QtWidgets.QLineEdit.TrailingPosition,
)

ActionPositionStr = Literal["leading", "trailing"]

QtWidgets.QLineEdit.__bases__ = (widgets.Widget,)


class LineEdit(QtWidgets.QLineEdit):

    value_changed = core.Signal(str)

    def __init__(
        self,
        default_value: str = "",
        read_only: bool = False,
        parent: Optional[QtWidgets.QWidget] = None,
    ):
        super().__init__(default_value, parent)
        self.textChanged.connect(self._set_validation_color)
        self.textChanged.connect(self.value_changed)
        self.set_read_only(read_only)

    def __repr__(self):
        return f"{type(self).__name__}: {self.serialize_fields()}"

    def __setstate__(self, state):
        super().__setstate__(state)
        self.set_text(state["text"])
        self.setValidator(state["validator"])
        self.setInputMask(state["input_mask"])
        self.setMaxLength(state["max_length"])
        self.setPlaceholderText(state["placeholder_text"])
        self.setReadOnly(state["read_only"])
        self.setFrame(state["has_frame"])
        self.setClearButtonEnabled(state["clear_button_enabled"])
        # self.setAlignment(state["alignment"])
        self.set_cursor_move_style(state["cursor_move_style"])
        self.set_echo_mode(state["echo_mode"])
        self.setCursorPosition(state["cursor_position"])
        self.setDragEnabled(state["drag_enabled"])
        self.setModified(state["is_modified"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __add__(self, other: str):
        self.append_text(other)
        return self

    def serialize_fields(self):
        return dict(
            text=self.text(),
            # alignment=self.alignment(),
            validator=self.validator(),
            max_length=self.maxLength(),
            read_only=self.isReadOnly(),
            input_mask=self.inputMask(),
            has_frame=self.hasFrame(),
            placeholder_text=self.placeholderText(),
            clear_button_enabled=self.isClearButtonEnabled(),
            cursor_move_style=self.get_cursor_move_style(),
            echo_mode=self.get_echo_mode(),
            cursor_position=self.cursorPosition(),
            drag_enabled=self.dragEnabled(),
            is_modified=self.isModified(),
        )

    def font(self) -> gui.Font:
        return gui.Font(super().font())

    def append_text(self, text: str):
        self.set_text(self.text() + text)

    def set_text(self, text: str):
        self.setText(text)

    def set_read_only(self, value: bool = True):
        """Set text to read-only.

        Args:
            value: True, for read-only, otherwise False
        """
        self.setReadOnly(value)

    def set_regex_validator(self, regex: str, flags=0) -> gui.RegularExpressionValidator:
        validator = gui.RegularExpressionValidator(self)
        validator.set_regex(regex, flags)
        self.set_validator(validator)
        return validator

    def set_range(self, lower: Optional[int], upper: Optional[int]):
        val = gui.IntValidator()
        val.set_range(lower, upper)
        self.set_validator(val)

    def set_validator(self, validator: gui.Validator):
        self.setValidator(validator)
        self._set_validation_color()

    def set_input_mask(self, mask: str):
        self.setInputMask(mask)

    def _set_validation_color(self, state: bool = True):
        color = "orange" if not self.is_valid() else None
        self.set_background_color(color)

    def set_echo_mode(self, mode: EchoModeStr):
        """Set echo mode.

        Args:
            mode: echo mode to use

        Raises:
            InvalidParamError: invalid echo mode
        """
        if mode not in ECHO_MODE:
            raise InvalidParamError(mode, ECHO_MODE)
        self.setEchoMode(ECHO_MODE[mode])

    def get_echo_mode(self) -> EchoModeStr:
        """Return echo mode.

        Returns:
            echo mode
        """
        return ECHO_MODE.inverse[self.echoMode()]

    def set_cursor_move_style(self, style: constants.CursorMoveStyleStr):
        """Set cursor move style.

        Args:
            style: cursor move style to use

        Raises:
            InvalidParamError: invalid cursor move style
        """
        if style not in constants.CURSOR_MOVE_STYLE:
            raise InvalidParamError(style, constants.CURSOR_MOVE_STYLE)
        self.setCursorMoveStyle(constants.CURSOR_MOVE_STYLE[style])

    def get_cursor_move_style(self) -> constants.CursorMoveStyleStr:
        """Return cursor move style.

        Returns:
            cursor move style
        """
        return constants.CURSOR_MOVE_STYLE.inverse[self.cursorMoveStyle()]

    def add_action(
        self, action: QtWidgets.QAction, position: ActionPositionStr = "trailing"
    ):
        self.addAction(action, ACTION_POSITION[position])

    def set_value(self, value: str):
        self.setText(value)

    def get_value(self) -> str:
        return self.text()

    def is_valid(self) -> bool:
        return self.hasAcceptableInput()


if __name__ == "__main__":
    app = widgets.app()
    widget = LineEdit()
    action = widgets.Action(text="hallo", icon="mdi.folder")
    widget.add_action(action)
    widget.setPlaceholderText("test")
    widget.setClearButtonEnabled(True)
    # widget.set_regex_validator("[0-9]+")
    widget.setFont(gui.Font("Consolas"))
    widget.show()
    app.main_loop()



# from prettyqt.qt import QtCore

# from prettyqt.utils import bidict, InvalidParamError

from __future__ import annotations


# VERBOSITY_LEVEL = bidict(
#     minimum=QtCore.QDebug.MinimumVerbosity,
#     default=QtCore.QDebug.DefaultVerbosity,
#     maximum=QtCore.QDebug.MaximumVerbosity,
# )

# class Debug(QtCore.QDebug):
#     def set_verbosity(self, verbosity: str):
#         """Set the debug verbosity.

#         Valid values are "minimum", "default", "maximum"

#         Args:
#             verbosity: debug verbosity

#         Raises:
#             InvalidParamError: invalid debug verbosity
#         """
#         if verbosity not in VERBOSITY_LEVEL:
#             raise InvalidParamError(verbosity, VERBOSITY_LEVEL)
#         self.setVerbosity(VERBOSITY_LEVEL[verbosity])

#     def get_verbosity(self) -> str:
#         """Get current debug verbosity.

#         Possible values are "minimum", "default", "maximum"

#         Returns:
#             current debug verbosity
#         """
#         return VERBOSITY_LEVEL.inverse[self.verbosity()]


# if __name__ == "__main__":
#     matcher = Debug("Test")
#     print(repr(matcher))



from __future__ import annotations

from prettyqt import positioning
from prettyqt.qt import QtPositioning


QtPositioning.QGeoRectangle.__bases__ = (positioning.GeoShape,)


class GeoRectangle(QtPositioning.QGeoRectangle):
    def __repr__(self):
        return (
            f"{type(self).__name__}({self.get_top_left()!r}, {self.get_bottom_right()!r})"
        )

    def get_top_left(self) -> positioning.GeoCoordinate:
        return positioning.GeoCoordinate(self.topLeft())

    def get_bottom_right(self) -> positioning.GeoCoordinate:
        return positioning.GeoCoordinate(self.bottomRight())


if __name__ == "__main__":
    coord1 = positioning.GeoCoordinate(1, 1)
    coord2 = positioning.GeoCoordinate(11, 11)
    rect = GeoRectangle(coord1, coord2)
    print(str(rect))
    print(repr(rect))



from __future__ import annotations

from typing import Any, Dict, Union

from prettyqt.qt import QtGui


class Color(QtGui.QColor):
    def __init__(self, *args):
        # PySide2 workaround
        if len(args) == 1:
            if isinstance(args[0], QtGui.QColor):
                super().__init__(args[0].name())
            elif isinstance(args[0], str):
                super().__init__()
                self.set_color(args[0])
            else:
                super().__init__(*args)
        else:
            super().__init__(*args)

    def __repr__(self):
        return (
            f"{type(self).__name__}({self.red()}, {self.green()}, "
            f"{self.blue()}, {self.alpha()})"
        )

    def __str__(self):
        return self.name()

    def __reduce__(self):
        return type(self), (self.red(), self.green(), self.blue(), self.alpha())

    def serialize_fields(self):
        return dict(color=self.toString())

    def serialize(self) -> Dict[str, Any]:
        return self.serialize_fields()

    def set_color(self, color: Union[str, tuple]):
        if isinstance(color, str):
            self.setNamedColor(color)
        else:
            self.setRgb(*color)

    @classmethod
    def from_text(cls, text: str) -> Color:
        """Create a QColor from specified string."""
        color = cls()
        if text.startswith("#") and len(text) == 7:
            correct = "#0123456789abcdef"
            for char in text:
                if char.lower() not in correct:
                    return color
        elif text not in list(cls.colorNames()):
            return color
        color.setNamedColor(text)
        return color

    @classmethod
    def from_cmyk(cls, *args, **kwargs) -> Color:
        return cls(cls.fromCmykF(*args, **kwargs))

    def as_qt(self) -> QtGui.QColor:
        return QtGui.QColor(self)



from __future__ import annotations

from prettyqt import core, custom_animations, widgets
from prettyqt.qt import QtCore, QtWidgets


class AnimatedToolTipEventFilter(QtWidgets.QAbstractButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tool_tip = widgets.Label()
        self.tool_tip.set_flags(frameless=True, tooltip=True)
        self.tool_tip.set_attributes(
            transparent_for_mouse_events=True,
            no_system_background=True,
            translucent_background=True,
        )
        self.tool_tip.hide()
        self.slide_anim = custom_animations.SlideAnimation(duration=1000)
        self.slide_anim.apply_to(self.tool_tip)
        self.fade_anim = custom_animations.FadeInAnimation(duration=1000)
        self.fade_anim.apply_to(self.tool_tip)
        self.animation = self.slide_anim | self.fade_anim

    def eventFilter(self, obj: QtWidgets.QAbstractButton, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.Enter:
            # self.tool_tip.adjustSize()
            self.tool_tip.setText(obj.toolTip())
            self.tool_tip.show()
            position = "bottom"
            distance = 100
            center = obj.rect().center()
            if position == "top":
                center.setY(obj.rect().bottom())
                delta = core.Point(0, distance)
            elif position == "bottom":
                center.setY(obj.rect().top())
                delta = core.Point(0, -distance)
            elif position == "left":
                center.setX(obj.rect().right())
                delta = core.Point(distance, 0)
            elif position == "right":
                center.setX(obj.rect().left())
                delta = core.Point(-distance, 0)
            center = obj.mapToGlobal(center)
            self.slide_anim.set_start_value(center + delta)
            self.slide_anim.set_end_value(center)
            self.animation.start_animation()
        elif event.type() == QtCore.QEvent.Leave:
            self.tool_tip.hide()
            self.animation.stop()
        elif event.type() == QtCore.QEvent.ToolTip:
            return True
        return super().eventFilter(obj, event)


if __name__ == "__main__":
    app = widgets.app()
    w = widgets.ToolButton()
    ef = AnimatedToolTipEventFilter()
    w.installEventFilter(ef)
    w.setText("test")
    w.set_icon("mdi.timer")
    w.set_tooltip("testus")
    w.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import constants, core, gui, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets


DecorationRole2 = QtCore.Qt.UserRole + 1000


class IconDelegate(widgets.StyledItemDelegate):
    def paint(
        self,
        painter: QtGui.QPainter,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ):
        """Override to paint an icon based on given Pixmap / Color / Icon.

        Pixmap / Color / Icon must be set to 'QtCore.Qt.UserRole + 1000'

        Args:
            painter (QtGui.QPainter): painter to paint the icon
            option (QtWidgets.QStyleOptionViewItem): state of the item to be displayed
            index (QtCore.QModelIndex): index which gets decorated
        """
        super().paint(painter, option, index)
        value = index.data(DecorationRole2)
        if not value:
            return
        margin = 10
        mode = gui.Icon.Normal

        if not (option.state & widgets.Style.State_Enabled):
            mode = gui.Icon.Disabled
        elif option.state & widgets.Style.State_Selected:
            mode = gui.Icon.Selected

        if isinstance(value, QtGui.QPixmap):
            icon = QtGui.QIcon(value)
            option.decorationSize = value.size() / value.devicePixelRatio()

        elif isinstance(value, QtGui.QColor):
            pixmap = QtGui.QPixmap(option.decorationSize)
            pixmap.fill(value)
            icon = QtGui.QIcon(pixmap)

        elif isinstance(value, QtGui.QImage):
            icon = QtGui.QIcon(QtGui.QPixmap.fromImage(value))
            option.decorationSize = value.size() / value.devicePixelRatio()

        elif isinstance(value, QtGui.QIcon):
            is_on = option.state & widgets.Style.State_Open
            state = gui.Icon.On if is_on else gui.Icon.Off
            actual_size = option.icon.actualSize(option.decorationSize, mode, state)
            option.decorationSize = option.decorationSize & actual_size
        r = core.Rect(core.Point(), option.decorationSize)
        r.moveCenter(option.rect.center())
        r.setRight(option.rect.right() - margin)
        state = gui.Icon.On if option.state & widgets.Style.State_Open else gui.Icon.Off
        alignment = constants.ALIGN_RIGHT | constants.ALIGN_V_CENTER
        icon.paint(painter, r, alignment, mode, state)



from __future__ import annotations

import contextlib

from prettyqt.qt import QtWidgets


class WhatsThis(QtWidgets.QWhatsThis):
    @classmethod
    @contextlib.contextmanager
    def enter_mode(cls):
        cls.enterWhatsThisMode()
        yield cls
        cls.leaveWhatsThisMode()


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()

    def test():
        with WhatsThis.enter_mode() as w:
            print(w.inWhatsThisMode())

    btn = widgets.PushButton(callback=test)
    btn.show()
    app.main_loop()



from __future__ import annotations

import os
import pathlib
from typing import Any, Dict, Literal, Optional, Union

from prettyqt.qt import QtCore
from prettyqt.utils import bidict


COMPONENT_FORMATTING_OPTIONS = bidict(
    pretty_decoded=QtCore.QUrl.PrettyDecoded,
    encode_spaces=QtCore.QUrl.EncodeSpaces,
    encode_unicode=QtCore.QUrl.EncodeUnicode,
    encode_delimiters=QtCore.QUrl.EncodeDelimiters,
    encode_reserved=QtCore.QUrl.EncodeReserved,
    decode_reserved=QtCore.QUrl.DecodeReserved,
    fully_encoded=QtCore.QUrl.FullyEncoded,
    fully_decoded=QtCore.QUrl.FullyDecoded,
)

ComponentFormattingStr = Literal[
    "pretty_decoded",
    "encode_spaces",
    "encode_unicode",
    "encode_delimiters",
    "encode_reserved",
    "decode_reserved",
    "fully_encoded",
    "fully_decoded",
]


PARSING_MODES = bidict(
    tolerant=QtCore.QUrl.TolerantMode,
    strict=QtCore.QUrl.StrictMode,
    decoded=QtCore.QUrl.DecodedMode,
)

ParsingModeStr = Literal["tolerant", "strict", "decoded"]

FORMATTING_OPTIONS = bidict(
    none=0,  # QtCore.QUrl.None
    remove_scheme=QtCore.QUrl.RemoveScheme,
    remove_password=QtCore.QUrl.RemovePassword,
    remove_user_info=QtCore.QUrl.RemoveUserInfo,
    remove_port=QtCore.QUrl.RemovePort,
    remove_authority=QtCore.QUrl.RemoveAuthority,
    remove_path=QtCore.QUrl.RemovePath,
    remove_query=QtCore.QUrl.RemoveQuery,
    remove_fragment=QtCore.QUrl.RemoveFragment,
    remove_filename=QtCore.QUrl.RemoveFilename,
    prefer_local_file=QtCore.QUrl.PreferLocalFile,
    strip_trailing_slash=QtCore.QUrl.StripTrailingSlash,
    normalize_path_segments=QtCore.QUrl.NormalizePathSegments,
)

FormattingOptionStr = Literal[
    "none",
    "remove_scheme",
    "remove_password",
    "remove_user_info",
    "remove_port",
    "remove_authority",
    "remove_path",
    "remove_query",
    "remove_fragment",
    "remove_filename",
    "prefer_local_file",
    "strip_trailing_slash",
    "normalize_path_segments",
]


class Url(QtCore.QUrl):
    def __init__(self, path: Optional[Union[QtCore.QUrl, str, os.PathLike]] = None):
        if path is None:
            super().__init__()
        else:
            if isinstance(path, QtCore.QUrl):
                super().__init__(path)
            else:
                super().__init__(os.fspath(path))
            if isinstance(path, os.PathLike):  # type: ignore
                self.setScheme("file")

    # def __str__(self):
    #     return self.absolutePath()

    def __repr__(self):
        return f"{type(self).__name__}({self.toString()!r})"

    def __str__(self):
        return self.toString()

    def serialize_fields(self):
        return dict(path=self.toString())

    def serialize(self) -> Dict[str, Any]:
        return self.serialize_fields()

    def to_string(self) -> str:
        return self.toString()

    def to_path(self) -> pathlib.Path:
        """Get pathlib object from the URL.

        Returns:
            Path
        """
        return pathlib.Path(str(self))

    def is_local_file(self) -> bool:
        return self.isLocalFile()

    @classmethod
    def from_user_input(cls, url: str, working_dir: Optional[str] = None) -> Url:
        if working_dir is None:
            working_dir = ""
        return cls(cls.fromUserInput(url, working_dir))


if __name__ == "__main__":
    url = Url()
    str(url)



from __future__ import annotations

from typing import Literal

from prettyqt import gui, widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError, bidict


CORRECTION_MODES = bidict(
    to_previous=QtWidgets.QSpinBox.CorrectToPreviousValue,
    to_nearest=QtWidgets.QSpinBox.CorrectToNearestValue,
)

CorrectionModeStr = Literal["to_previous", "to_nearest"]

SYMBOLS = bidict(
    up_down=QtWidgets.QSpinBox.UpDownArrows,
    plus_minus=QtWidgets.QSpinBox.PlusMinus,
    none=QtWidgets.QSpinBox.NoButtons,
)

SymbolStr = Literal["up_down", "plus_minus", "none"]

STEP_TYPES = bidict(
    default=QtWidgets.QSpinBox.DefaultStepType,
    adaptive=QtWidgets.QSpinBox.AdaptiveDecimalStepType,
)

StepTypeStr = Literal["default", "adaptive"]


QtWidgets.QAbstractSpinBox.__bases__ = (widgets.Widget,)


class AbstractSpinBox(QtWidgets.QAbstractSpinBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setLineEdit(widgets.LineEdit())
        self.setGroupSeparatorShown(True)

    def serialize_fields(self):
        return dict(
            button_symbols=self.get_button_symbols(),
            correction_mode=self.get_correction_mode(),
        )

    def __setstate__(self, state):
        super().__setstate__(state)
        self.set_correction_mode(state["correction_mode"])
        self.set_button_symbols(state["button_symbols"])

    def is_valid(self) -> bool:
        return self.hasAcceptableInput()

    def set_validator(self, validator: gui.Validator):
        self.lineEdit().setValidator(validator)

    def get_button_symbols(self) -> SymbolStr:
        """Return button symbol type.

        Returns:
            button symbol type
        """
        return SYMBOLS.inverse[self.buttonSymbols()]

    def set_button_symbols(self, mode: SymbolStr):
        """Set button symbol type.

        Args:
            mode: button symbol type to use

        Raises:
            InvalidParamError: invalid button symbol type
        """
        if mode not in SYMBOLS:
            raise InvalidParamError(mode, SYMBOLS)
        self.setButtonSymbols(SYMBOLS[mode])

    def set_correction_mode(self, mode: CorrectionModeStr):
        """Set correction mode.

        Args:
            mode: correction mode to use

        Raises:
            InvalidParamError: invalid correction mode
        """
        if mode not in CORRECTION_MODES:
            raise InvalidParamError(mode, CORRECTION_MODES)
        self.setCorrectionMode(CORRECTION_MODES[mode])

    def get_correction_mode(self) -> CorrectionModeStr:
        """Return correction mode.

        Returns:
            correction mode
        """
        return CORRECTION_MODES.inverse[self.correctionMode()]

    def set_step_type(self, mode: StepTypeStr):
        """Set step type.

        Args:
            mode: step type to use

        Raises:
            InvalidParamError: invalid step type
        """
        if mode not in STEP_TYPES:
            raise InvalidParamError(mode, STEP_TYPES)
        self.setStepType(STEP_TYPES[mode])

    def get_step_type(self) -> StepTypeStr:
        """Return step type.

        Returns:
            step type
        """
        return STEP_TYPES.inverse[self.stepType()]

    def set_special_value(self, value: str):
        self.setSpecialValueText(value)

    def get_value(self) -> int:
        return self.value()

    def set_value(self, value: int):
        self.setValue(value)


if __name__ == "__main__":
    app = widgets.app()
    widget = AbstractSpinBox()
    widget.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QStyleOptionTabBarBase.__bases__ = (widgets.StyleOption,)


class StyleOptionTabBarBase(QtWidgets.QStyleOptionTabBarBase):
    pass



# see https://github.com/ITVRoC/SeekurJr/blob/master/seekur_12.04/packages/
# multimaster_fkie/node_manager_fkie/src/node_manager_fkie/yaml_highlighter.py

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional, Union

import regex as re

from prettyqt import gui


@dataclass
class HighlightRule:
    regex: Union[str, List[str]] = ""
    color: str = "black"
    italic: bool = False
    bold: bool = False
    minimal: bool = False
    font_size: Optional[float] = None
    nth: int = 0
    compiled = None
    fmt: gui.TextCharFormat = gui.TextCharFormat()

    def __init_subclass__(cls):
        super().__init_subclass__()
        if isinstance(cls.regex, str):
            cls.compiled = re.compile(cls.regex)
            # cls.compiled.setMinimal(True)
        else:
            cls.compiled = [re.compile(r) for r in cls.regex]
        cls.fmt = cls.get_format()

    @classmethod
    def get_format(cls) -> gui.TextCharFormat:
        fmt = gui.TextCharFormat()
        fmt.setFontItalic(cls.italic)
        fmt.set_foreground_color(cls.color)
        if cls.font_size:
            fmt.setFontPointSize(cls.font_size)
        if cls.bold:
            fmt.set_font_weight("bold")
        return fmt



from __future__ import annotations

from typing import Any, Iterator, List, Union

import orjson as json

from prettyqt import core
from prettyqt.qt import QtCore
from prettyqt.utils import types


QtCore.QMimeData.__bases__ = (core.Object,)
OPTS = json.OPT_NAIVE_UTC | json.OPT_SERIALIZE_NUMPY


class MimeData(QtCore.QMimeData):
    def __len__(self):
        return len(self.formats())

    def __getitem__(self, index: str) -> str:
        return self.get_data(index)

    def __setitem__(self, index: str, value: Union[QtCore.QByteArray, bytes, str]):
        if isinstance(value, str):
            value = value.encode()
        if not isinstance(value, QtCore.QByteArray):
            value = QtCore.QByteArray(value)
        self.setData(index, value)

    def __contains__(self, fmt: str):
        return self.hasFormat(fmt)

    def __delitem__(self, index: str):
        self.removeFormat(index)

    def set_data(self, mime_type: str, data: str):
        self.setData(mime_type, QtCore.QByteArray(data.encode()))

    def set_json_data(self, mime_type: str, data: types.JSONType):
        self.setData(mime_type, QtCore.QByteArray(json.dumps(data, option=OPTS)))

    def get_data(self, mime_type: str) -> str:
        return bytes(self.data(mime_type)).decode()

    def get_json_data(self, mime_type: str) -> types.JSONType:
        data = self.data(mime_type)
        return json.loads(bytes(data))

    def keys(self) -> List[str]:
        return self.formats()

    def values(self) -> Iterator[Any]:
        return (self.get_data(key) for key in self.formats())


if __name__ == "__main__":
    mime_data = MimeData()



from __future__ import annotations

from prettyqt import positioning
from prettyqt.qt import QtPositioning


class GeoLocation(QtPositioning.QGeoLocation):
    def get_address(self) -> positioning.GeoAddress:
        return positioning.GeoAddress(self.address())

    def get_coordinate(self) -> positioning.GeoCoordinate:
        return positioning.GeoCoordinate(self.coordinate())

    def get_bounding_box(self) -> positioning.GeoRectangle:
        return positioning.GeoRectangle(self.boundingBox())



from __future__ import annotations

from typing import Optional, Tuple, Union

from prettyqt import core
from prettyqt.qt import QtCore, QtGui


QtGui.QAbstractTextDocumentLayout.__bases__ = (core.Object,)


class AbstractTextDocumentLayout(QtGui.QAbstractTextDocumentLayout):
    def __repr__(self):
        return f"{type(self).__name__}()"

    def __len__(self):
        return self.pageCount()

    def get_block_bounding_rect(self, block: QtGui.QTextBlock) -> core.RectF:
        return core.RectF(self.blockBoundingRect(block))

    def get_frame_bounding_rect(self, frame: QtGui.QTextBlock) -> core.RectF:
        return core.RectF(self.frameBoundingRect(frame))

    def hit_test(
        self, point: Union[core.PointF, Tuple[float, float]], fuzzy: bool = False
    ) -> Optional[int]:
        if isinstance(point, tuple):
            point = core.PointF(*point)
        accuracy = QtCore.Qt.FuzzyHit if fuzzy else QtCore.Qt.ExactHit
        result = self.hitTest(point, accuracy)
        if result == -1:
            return None
        return result


if __name__ == "__main__":
    from prettyqt import gui

    doc = gui.TextDocument()
    layout = AbstractTextDocumentLayout(doc)



from __future__ import annotations

import contextlib
from typing import Iterator, List, Literal, Union

from prettyqt import constants, core, gui
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import InvalidParamError, bidict, colors, types


COMPOSITION_MODE = bidict(
    source_over=QtGui.QPainter.CompositionMode_SourceOver,
    destination_over=QtGui.QPainter.CompositionMode_DestinationOver,
    clear=QtGui.QPainter.CompositionMode_Clear,
    source=QtGui.QPainter.CompositionMode_Source,
    destination=QtGui.QPainter.CompositionMode_Destination,
    source_in=QtGui.QPainter.CompositionMode_SourceIn,
    destination_in=QtGui.QPainter.CompositionMode_DestinationIn,
    source_out=QtGui.QPainter.CompositionMode_SourceOut,
    destination_out=QtGui.QPainter.CompositionMode_DestinationOut,
    source_atop=QtGui.QPainter.CompositionMode_SourceAtop,
    destination_atop=QtGui.QPainter.CompositionMode_DestinationAtop,
)

CompositionModeStr = Literal[
    "source_over",
    "destination_over",
    "clear",
    "source",
    "destination",
    "source_in",
    "destination_in",
    "source_out",
    "destination_out",
    "source_atop",
    "destination_atop",
]

RENDER_HINTS = bidict(
    antialiasing=QtGui.QPainter.Antialiasing,
    text_antialiasing=QtGui.QPainter.TextAntialiasing,
    smooth_pixmap_transform=QtGui.QPainter.SmoothPixmapTransform,
    lossless_image_rendering=QtGui.QPainter.LosslessImageRendering,
)

RenderHintStr = Literal[
    "antialiasing",
    "text_antialiasing",
    "smooth_pixmap_transform",
    "lossless_image_rendering",
]


class Painter(QtGui.QPainter):
    def __enter__(self):
        return self

    def __exit__(self, typ, value, traceback):
        self.end()

    @contextlib.contextmanager
    def paint_on(self, obj) -> Iterator[Painter]:
        self.begin(obj)
        yield self
        self.end()

    @contextlib.contextmanager
    def backup_state(self) -> Iterator[Painter]:
        self.save()
        yield self
        self.restore()

    @contextlib.contextmanager
    def native_mode(self) -> Iterator[Painter]:
        self.beginNativePainting()
        yield self
        self.endNativePainting()

    @contextlib.contextmanager
    def edit_pen(self) -> Iterator[gui.Pen]:
        pen = gui.Pen(self.pen())
        yield pen
        self.setPen(pen)

    def draw_image(
        self,
        target: Union[QtCore.QPoint, QtCore.QPointF, QtCore.QRect, QtCore.QRectF],
        frame_buffer: QtGui.QImage,
    ):
        self.set_composition_mode("source_atop")
        self.drawImage(target, frame_buffer)

    def draw_polygon(
        self,
        points: Union[
            QtGui.QPolygon, QtGui.QPolygonF, List[QtCore.QPoint], List[QtCore.QPointF]
        ],
        fill_rule: constants.FillRuleStr = "odd_even",
    ):
        if fill_rule not in constants.FILL_RULE:
            raise InvalidParamError(fill_rule, constants.FILL_RULE)
        self.drawPolygon(points, fillRule=constants.FILL_RULE[fill_rule])  # type: ignore

    def use_antialiasing(self):
        self.setRenderHint(self.Antialiasing, True)

    def fill_rect(
        self,
        rect: Union[QtCore.QRectF, QtCore.QRect],
        color: types.ColorType,
        pattern: constants.PatternStr = "solid",
    ):
        if pattern not in constants.PATTERN:
            raise InvalidParamError(pattern, constants.PATTERN)
        if isinstance(rect, tuple):
            rect = core.Rect(*rect)
        color = colors.get_color(color)
        if pattern != "solid":
            color = gui.Brush(color, constants.PATTERN[pattern])
        self.fillRect(rect, color)

    def set_pen(
        self,
        style: constants.PenStyleStr = "solid",
        width: float = 1.0,
        color: types.ColorType = "black",
        join_style: constants.JoinStyleStr = "bevel",
        cap_style: constants.CapStyleStr = "square",
    ) -> gui.Pen:
        """Set pen to use.

        Args:
            style: pen style to use
            width: pen width
            color: pen color
            join_style: pen join style to use
            cap_style: pen cap style to use
        """
        pen = gui.Pen()
        pen.set_style(style)
        pen.set_cap_style(cap_style)
        pen.set_join_style(join_style)
        pen.setWidthF(width)
        pen.set_color(color)
        self.setPen(pen)
        return pen

    def get_pen(self) -> gui.Pen:
        """Return current pen.

        Returns:
            current pen
        """
        return gui.Pen(self.pen())

    def set_color(self, color: types.ColorType):
        color = colors.get_color(color)
        self.setPen(color)

    def set_brush(self, brush: Union[QtGui.QBrush, types.ColorType]):
        if not isinstance(brush, QtGui.QBrush):
            brush = colors.get_color(brush)
        self.setBrush(brush)

    def set_transparent_background(self, transparent: bool = True):
        mode = QtCore.Qt.TransparentMode if transparent else QtCore.Qt.OpaqueMode
        self.setBackgroundMode(mode)

    def set_composition_mode(self, mode: CompositionModeStr):
        """Set the current composition mode.

        Raises:
            InvalidParamError: composition mode does not exist
        """
        if mode not in COMPOSITION_MODE:
            raise InvalidParamError(mode, COMPOSITION_MODE)
        self.setCompositionMode(COMPOSITION_MODE[mode])

    def get_composition_mode(self) -> CompositionModeStr:
        """Get the current composition mode.

        Returns:
            composition mode
        """
        return COMPOSITION_MODE.inverse[self.compositionMode()]

    def set_clip_path(
        self, path: QtGui.QPainterPath, operation: constants.ClipOperationStr = "replace"
    ):
        if operation not in constants.CLIP_OPERATION:
            raise InvalidParamError(operation, constants.CLIP_OPERATION)
        self.setClipPath(path, constants.CLIP_OPERATION[operation])

    def get_text_rect(self, text: str) -> core.Rect:
        return self.drawText(core.Rect(), QtCore.Qt.TextDontPrint, text)  # type: ignore

    @contextlib.contextmanager
    def clip_path(
        self, operation: constants.ClipOperationStr = "replace"
    ) -> Iterator[gui.PainterPath]:
        path = gui.PainterPath()
        yield path
        self.set_clip_path(path, operation)

    @contextlib.contextmanager
    def apply_transform(self, combine: bool = True) -> Iterator[gui.Transform]:
        transform = gui.Transform()
        yield transform
        self.setTransform(transform, combine)

    @contextlib.contextmanager
    def offset_by(self, x: int = 0, y: int = 0) -> Iterator[Painter]:
        self.translate(x, y)
        yield self
        self.translate(-x, -y)


if __name__ == "__main__":
    painter = Painter()



from __future__ import annotations

from typing import Iterator, List, Optional, Union

from prettyqt import gui, iconprovider, widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import types


QtWidgets.QToolBox.__bases__ = (widgets.Frame,)


class ToolBox(QtWidgets.QToolBox):
    def __getitem__(self, index: Union[int, str]) -> QtWidgets.QWidget:
        if isinstance(index, int):
            return self.widget(index)
        else:
            result = self.find_child(QtWidgets.QWidget, index)
            if result is None:
                raise KeyError("Widget not found")
            return result

    def __delitem__(self, index: int):
        self.removeItem(index)

    def serialize_fields(self):
        children = list()
        for i, widget in enumerate(self.get_children()):
            dct = dict(
                widget=widget,
                icon=self.item_icon(i),
                text=self.itemText(i),
                enabled=self.isItemEnabled(i),
                tool_tip=self.itemToolTip(i),
            )
            children.append(dct)
        return dict(items=children, current_index=self.currentIndex())

    def __setstate__(self, state):
        for i, item in enumerate(state["items"]):
            self.addItem(item["widget"], item["icon"], item["text"])
            self.setItemEnabled(i, item["enabled"])
            self.setItemToolTip(i, item["tool_tip"])
        self.setCurrentIndex(state["current_index"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __iter__(self) -> Iterator[QtWidgets.QWidget]:
        return iter(self.get_children())

    def __contains__(self, item: QtWidgets.QWidget):
        return self.indexOf(item) >= 0

    def get_children(self) -> List[QtWidgets.QWidget]:
        return [self[i] for i in range(self.count())]

    def add_widget(
        self,
        widget: QtWidgets.QWidget,
        title: Optional[str] = None,
        icon: types.IconType = None,
    ):
        if title is None:
            title = widget.objectName()
        if icon:
            icon = iconprovider.get_icon(icon)
            self.addItem(widget, icon, title)
        else:
            self.addItem(widget, title)

    def item_icon(self, index: int) -> Optional[gui.Icon]:
        icon = self.itemIcon(index)
        if icon.isNull():
            return None
        return gui.Icon(icon)


if __name__ == "__main__":
    app = widgets.app()
    w = widgets.Widget()
    w2 = widgets.Widget()
    tb = ToolBox()
    tb.add_widget(w, "title")
    tb.add_widget(w2)
    print(len(tb))
    tb.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import charts
from prettyqt.qt.QtCharts import QtCharts


QtCharts.QPercentBarSeries.__bases__ = (charts.AbstractBarSeries,)


class PercentBarSeries(QtCharts.QPercentBarSeries):
    pass



"""Gui module.

Contains QtGui-based classes
"""
from __future__ import annotations

from prettyqt.qt.QtGui import (
    QKeyEvent as KeyEvent,
    QMouseEvent as MouseEvent,
    QWheelEvent as WheelEvent,
    QActionEvent as ActionEvent,
    QShowEvent as ShowEvent,
    QDropEvent as DropEvent,
    QFocusEvent as FocusEvent,
    QDragEnterEvent as DragEnterEvent,
    QContextMenuEvent as ContextMenuEvent,
    QResizeEvent as ResizeEvent,
    QNativeGestureEvent as NativeGestureEvent,
    QMoveEvent as MoveEvent,
    QInputMethodEvent as InputMethodEvent,
    QInputMethodQueryEvent as InputMethodQueryEvent,
    QCloseEvent as CloseEvent,
    QDragLeaveEvent as DragLeaveEvent,
    QHelpEvent as HelpEvent,
    QHideEvent as HideEvent,
    QHoverEvent as HoverEvent,
    QDragMoveEvent as DragMoveEvent,
    QEnterEvent as EnterEvent,
    QExposeEvent as ExposeEvent,
    QFileOpenEvent as FileOpenEvent,
    QIconDragEvent as IconDragEvent,
    QInputEvent as InputEvent,
    QPaintEvent as PaintEvent,
    QShortcutEvent as ShortcutEvent,
    QStatusTipEvent as StatusTipEvent,
    QTouchEvent as TouchEvent,
    QTabletEvent as TabletEvent,
    QWindowStateChangeEvent as WindowStateChangeEvent,
    QWhatsThisClickedEvent as WhatsThisClickedEvent,
    QScrollEvent as ScrollEvent,
    QScrollPrepareEvent as ScrollPrepareEvent,
)

# not available in PySide2
# from prettyqt.qt.QtGui import QPlatformSurfaceEvent as PlatformSurfaceEvent

from .textobjectinterface import TextObjectInterface
from .screen import Screen
from .keysequence import KeySequence
from .surface import Surface
from .window import Window
from .icon import Icon
from .paintdevice import PaintDevice
from .paintdevicewindow import PaintDeviceWindow
from .pixmap import Pixmap
from .iconengine import IconEngine
from .pixmapcache import PixmapCache
from .bitmap import Bitmap
from .image import Image

from .rasterwindow import RasterWindow
from .clipboard import Clipboard
from .inputmethod import InputMethod
from .sessionmanager import SessionManager
from .validator import Validator

from .regularexpressionvalidator import RegularExpressionValidator
from .textline import TextLine
from .textoption import TextOption
from .textblock import TextBlock
from .textdocument import TextDocument
from .statictext import StaticText
from .intvalidator import IntValidator
from .doublevalidator import DoubleValidator
from .color import Color
from .brush import Brush
from .textblockuserdata import TextBlockUserData
from .gradient import Gradient
from .lineargradient import LinearGradient
from .radialgradient import RadialGradient
from .conicalgradient import ConicalGradient
from .pagesize import PageSize
from .pagelayout import PageLayout
from .font import Font
from .fontmetrics import FontMetrics
from .fontmetricsf import FontMetricsF
from .fontinfo import FontInfo
from .fontdatabase import FontDatabase
from .region import Region
from .movie import Movie
from .transform import Transform
from .pagedpaintdevice import PagedPaintDevice
from .pen import Pen
from .picture import Picture
from .painter import Painter
from .painterpath import PainterPath
from .painterpathstroker import PainterPathStroker
from .palette import Palette
from .guiapplication import GuiApplication
from .cursor import Cursor
from .polygon import Polygon
from .polygonf import PolygonF
from .standarditem import StandardItem
from .standarditemmodel import StandardItemModel
from .textcursor import TextCursor
from .pdfwriter import PdfWriter
from .desktopservices import DesktopServices
from .matrix4x4 import Matrix4x4
from .vector3d import Vector3D
from .vector4d import Vector4D
from .imageiohandler import ImageIOHandler
from .imagereader import ImageReader
from .imagewriter import ImageWriter
from .textformat import TextFormat
from .textlength import TextLength
from .textframeformat import TextFrameFormat
from .textcharformat import TextCharFormat
from .textimageformat import TextImageFormat
from .textlistformat import TextListFormat
from .texttablecellformat import TextTableCellFormat
from .textobject import TextObject
from .textblockgroup import TextBlockGroup
from .textframe import TextFrame
from .abstracttextdocumentlayout import AbstractTextDocumentLayout
from .syntaxhighlighter import SyntaxHighlighter

from prettyqt import core

if core.VersionNumber.get_qt_version() >= (5, 13, 0):
    from .colorspace import ColorSpace


def app():
    if GuiApplication.instance() is not None:
        return GuiApplication.instance()
    return GuiApplication([])


__all__ = [
    "app",
    "KeyEvent",
    "MouseEvent",
    "WheelEvent",
    "ActionEvent",
    "ShowEvent",
    "ContextMenuEvent",
    "ResizeEvent",
    "NativeGestureEvent",
    "InputMethodQueryEvent",
    "InputMethodEvent",
    "TextObjectInterface",
    "SessionManager",
    "CloseEvent",
    "DragLeaveEvent",
    "MoveEvent",
    "HelpEvent",
    "HideEvent",
    "HoverEvent",
    "DragMoveEvent",
    "TouchEvent",
    "TabletEvent",
    "WindowStateChangeEvent",
    "FileOpenEvent",
    "IconDragEvent",
    "InputEvent",
    "PaintEvent",
    "ShortcutEvent",
    "StatusTipEvent",
    "EnterEvent",
    "ExposeEvent",
    "WhatsThisClickedEvent",
    "ScrollEvent",
    "ScrollPrepareEvent",
    "PlatformSurfaceEvent",
    "DropEvent",
    "DragEnterEvent",
    "FocusEvent",
    "Screen",
    "Gradient",
    "LinearGradient",
    "RadialGradient",
    "ConicalGradient",
    "PageSize",
    "PageLayout",
    "Clipboard",
    "GuiApplication",
    "Validator",
    "TextLine",
    "TextOption",
    "TextBlock",
    "TextBlockGroup",
    "TextDocument",
    "StaticText",
    "RegularExpressionValidator",
    "IntValidator",
    "DoubleValidator",
    "Brush",
    "TextBlockUserData",
    "Color",
    "Font",
    "FontMetrics",
    "FontMetricsF",
    "FontInfo",
    "FontDatabase",
    "Region",
    "Movie",
    "PaintDevice",
    "Transform",
    "PagedPaintDevice",
    "Icon",
    "Image",
    "Pen",
    "Picture",
    "Pixmap",
    "IconEngine",
    "PixmapCache",
    "Bitmap",
    "Painter",
    "PainterPath",
    "PainterPathStroker",
    "Palette",
    "Cursor",
    "Polygon",
    "PolygonF",
    "StandardItem",
    "StandardItemModel",
    "TextCharFormat",
    "TextImageFormat",
    "TextListFormat",
    "TextTableCellFormat",
    "TextCursor",
    "SyntaxHighlighter",
    "PdfWriter",
    "KeySequence",
    "Surface",
    "Window",
    "DesktopServices",
    "Matrix4x4",
    "Vector3D",
    "Vector4D",
    "PaintDeviceWindow",
    "RasterWindow",
    "ImageIOHandler",
    "ImageReader",
    "ImageWriter",
    "TextObject",
    "TextLength",
    "TextFormat",
    "TextFrameFormat",
    "TextFrame",
    "AbstractTextDocumentLayout",
    "InputMethod",
    "ColorSpace",
]



from __future__ import annotations

from prettyqt import gui
from prettyqt.qt import QtWidgets


QtWidgets.QPlainTextDocumentLayout.__bases__ = (gui.AbstractTextDocumentLayout,)


class PlainTextDocumentLayout(QtWidgets.QPlainTextDocumentLayout):
    def serialize_fields(self):
        return dict(cursor_width=self.cursorWidth())


if __name__ == "__main__":
    doc = gui.TextDocument()
    layout = PlainTextDocumentLayout(doc)
    print(len(layout))



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtHelp


QtHelp.QHelpContentModel.__bases__ = (core.AbstractItemModel,)


class HelpContentModel(QtHelp.QHelpContentModel):
    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtGui


class Vector4D(QtGui.QVector4D):
    # def __repr__(self):
    #     return f"{type(self).__name__}()"
    pass

    def __bool__(self):
        return not self.isNull()

    def __abs__(self) -> float:
        return self.length()

    def to_point(self) -> core.Point:
        return core.Point(self.toPoint())

    def to_pointf(self) -> core.PointF:
        return core.PointF(self.toPointF())


if __name__ == "__main__":
    vector = Vector4D(0, 0, 0, 1)
    print(abs(vector))



from __future__ import annotations

from prettyqt import gui, widgets
from prettyqt.qt import QtWidgets


QtWidgets.QGraphicsOpacityEffect.__bases__ = (widgets.GraphicsEffect,)


class GraphicsOpacityEffect(QtWidgets.QGraphicsOpacityEffect):
    def serialize_fields(self):
        return dict(opacity=self.opacity(), opacity_mask=self.get_opacity_mask())

    def __setstate__(self, state):
        self.setOpacity(state["opacity"])
        self.setOpacityMask(state["opacity_mask"])

    def get_opacity_mask(self) -> gui.Brush:
        return gui.Brush(self.opacityMask())



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QStyleOptionDockWidget.__bases__ = (widgets.StyleOption,)


class StyleOptionDockWidget(QtWidgets.QStyleOptionDockWidget):
    pass



from __future__ import annotations

from typing import Literal, Optional, Union

from prettyqt import constants, core, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, bidict


TICK_POSITION = bidict(
    none=QtWidgets.QSlider.NoTicks,
    both_sides=QtWidgets.QSlider.TicksBothSides,
    above=QtWidgets.QSlider.TicksAbove,
    below=QtWidgets.QSlider.TicksBelow,
)

TickPositionAllStr = Literal["none", "both_sides", "above", "below", "left", "right"]
TickPositionStr = Literal["none", "both_sides", "above", "below"]


QtWidgets.QSlider.__bases__ = (widgets.AbstractSlider,)


class Slider(QtWidgets.QSlider):

    value_changed = core.Signal(int)

    def __init__(
        self,
        orientation: Union[
            constants.OrientationStr, QtCore.Qt.Orientation
        ] = "horizontal",
        parent: Optional[QtWidgets.QWidget] = None,
    ):
        if isinstance(orientation, QtCore.Qt.Orientation):
            ori = orientation
        else:
            ori = constants.ORIENTATION[orientation]
        super().__init__(ori, parent)
        self.valueChanged.connect(self.on_value_change)

    def serialize_fields(self):
        return dict(
            tick_position=self.get_tick_position(),
            tick_interval=self.tickInterval(),
        )

    def __setstate__(self, state):
        super().__setstate__(state)
        self.set_tick_position(state["tick_position"])
        self.setTickInterval(state["tick_interval"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def set_tick_position(self, position: TickPositionAllStr):
        """Set the tick position for the slider.

        For vertical orientation, "above" equals to "left" and "below" to "right".

        Args:
            position: position for the ticks
        """
        if position == "left":
            position = "above"
        elif position == "right":
            position = "below"
        elif position not in TICK_POSITION:
            raise InvalidParamError(position, TICK_POSITION)
        self.setTickPosition(TICK_POSITION[position])

    def get_tick_position(self) -> TickPositionStr:
        """Return tick position.

        Returns:
            tick position
        """
        return TICK_POSITION.inverse[self.tickPosition()]
        # if self.is_vertical():
        #     if val == "above":
        #         return "left"
        #     elif val == "below":
        #         return "right"


if __name__ == "__main__":
    app = widgets.app()
    slider = Slider()
    slider.setRange(0, 100)
    slider.value_changed.connect(print)
    slider.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QEventTransition.__bases__ = (core.AbstractTransition,)


class EventTransition(QtCore.QEventTransition):
    pass



from __future__ import annotations

from typing import Any, Dict

from prettyqt import constants, core
from prettyqt.qt import QtGui
from prettyqt.utils import InvalidParamError


class Cursor(QtGui.QCursor):
    def __getstate__(self):
        return bytes(self)

    def __setstate__(self, ba):
        super().__init__()
        core.DataStream.write_bytearray(ba, self)

    def __bytes__(self):
        ba = core.DataStream.create_bytearray(self)
        return bytes(ba)

    def serialize_fields(self) -> Dict[str, Any]:
        return dict(shape=self.get_shape())

    def set_shape(self, shape: constants.CursorShapeStr):
        """Set cursor shape.

        Args:
            shape: shape to use

        Raises:
            InvalidParamError: shape does not exist
        """
        if shape not in constants.CURSOR_SHAPE:
            raise InvalidParamError(shape, constants.CURSOR_SHAPE)
        self.setShape(constants.CURSOR_SHAPE[shape])

    def get_shape(self) -> constants.CursorShapeStr:
        """Return current cursor shape.

        Returns:
            cursor shape
        """
        return constants.CURSOR_SHAPE.inverse[self.shape()]

    @staticmethod
    def get_position() -> core.Point:
        return core.Point(Cursor.pos())



from __future__ import annotations

from prettyqt import location
from prettyqt.qt import QtLocation


QtLocation.QPlaceProposedSearchResult.__bases__ = (location.PlaceSearchResult,)


class PlaceProposedSearchResult(QtLocation.QPlaceProposedSearchResult):
    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtHelp


QtHelp.QHelpIndexModel.__bases__ = (core.StringListModel,)


class HelpIndexModel(QtHelp.QHelpIndexModel):
    pass



from __future__ import annotations

from typing import Callable, Iterator, Tuple, Union

from prettyqt import core
from prettyqt.qt import QtCore
from prettyqt.utils import types


QtCore.QVariantAnimation.__bases__ = (core.AbstractAnimation,)


class VariantAnimation(QtCore.QVariantAnimation):
    def __getitem__(self, value: float) -> types.Variant:
        return self.keyValueAt(value)

    def __setitem__(self, key: float, value: types.Variant):
        self.setKeyValueAt(key, value)

    def __iter__(self) -> Iterator[Tuple[float, types.Variant]]:
        return iter(self.keyValues())

    def serialize_fields(self):
        return dict(
            duration=self.duration(),
            easing_curve=self.get_easing(),
            key_values=self.keyValues(),
        )

    def set_easing(
        self, easing_type: Union[core.easingcurve.TypeStr, Callable[[float], float]]
    ):
        curve = core.EasingCurve()
        if isinstance(easing_type, str):
            curve.set_type(easing_type)
        else:
            curve.set_custom_type(easing_type)
        self.setEasingCurve(curve)

    def get_easing(self) -> Union[core.easingcurve.TypeStr, Callable[[float], float]]:
        curve = core.EasingCurve(self.easingCurve())
        typ = curve.get_type()
        if typ == "custom":
            return curve.get_custom_type()
        else:
            return typ

    def set_range(self, start, end):
        self.setStartValue(start)
        self.setEndValue(end)



from __future__ import annotations

from typing import Literal

from prettyqt import constants, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, bidict


EDGE = bidict(
    left=QtCore.Qt.AnchorLeft,
    horizontal_center=QtCore.Qt.AnchorHorizontalCenter,
    right=QtCore.Qt.AnchorRight,
    top=QtCore.Qt.AnchorTop,
    vertical_center=QtCore.Qt.AnchorVerticalCenter,
    bottom=QtCore.Qt.AnchorBottom,
)

EdgeStr = Literal[
    "left", "horizontal_center", "right", "top", "vertical_center", "bottom"
]

QtWidgets.QGraphicsAnchorLayout.__bases__ = (widgets.GraphicsLayout,)


class GraphicsAnchorLayout(QtWidgets.QGraphicsAnchorLayout):
    def add_anchor(
        self,
        first_item: QtWidgets.QGraphicsLayoutItem,
        first_edge: EdgeStr,
        second_item: QtWidgets.QGraphicsLayoutItem,
        second_edge: EdgeStr,
    ) -> QtWidgets.QGraphicsAnchor:
        return self.addAnchor(
            first_item, EDGE[first_edge], second_item, EDGE[second_edge]
        )

    def get_anchor(
        self,
        first_item: QtWidgets.QGraphicsLayoutItem,
        first_edge: EdgeStr,
        second_item: QtWidgets.QGraphicsLayoutItem,
        second_edge: EdgeStr,
    ) -> QtWidgets.QGraphicsAnchor:
        return self.anchor(first_item, EDGE[first_edge], second_item, EDGE[second_edge])

    def add_anchors(
        self,
        first_item: QtWidgets.QGraphicsLayoutItem,
        second_item: QtWidgets.QGraphicsLayoutItem,
        orientation: constants.OrientationStr,
    ):
        if orientation not in constants.ORIENTATION:
            raise InvalidParamError(orientation, constants.ORIENTATION)
        self.addAnchors(
            first_item, second_item, constants.ORIENTATION[orientation]  # type: ignore
        )

    def add_corner_anchors(
        self,
        first_item: QtWidgets.QGraphicsLayoutItem,
        first_corner: constants.CornerStr,
        second_item: QtWidgets.QGraphicsLayoutItem,
        second_corner: constants.CornerStr,
    ):
        self.addCornerAnchors(
            first_item,
            constants.CORNER[first_corner],
            second_item,
            constants.CORNER[second_corner],
        )


if __name__ == "__main__":
    app = widgets.app()
    layout = GraphicsAnchorLayout()



from __future__ import annotations

from prettyqt import location
from prettyqt.qt import QtLocation


QtLocation.QPlaceEditorial.__bases__ = (location.PlaceContent,)


class PlaceEditorial(QtLocation.QPlaceEditorial):
    def __str__(self):
        return f"{self.title()}: {self.text()}"


if __name__ == "__main__":
    editorial = PlaceEditorial()



from __future__ import annotations

from prettyqt import core, gui, widgets
from prettyqt.qt import QtGui, QtWidgets


QtWidgets.QKeySequenceEdit.__bases__ = (widgets.Widget,)


class KeySequenceEdit(QtWidgets.QKeySequenceEdit):

    value_changed = core.Signal(QtGui.QKeySequence)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.keySequenceChanged.connect(self.value_changed)

    def __repr__(self):
        return f"{type(self).__name__}({self.get_value()!r})"

    def set_value(self, value: str):
        seq = gui.KeySequence.fromString(value)
        self.setKeySequence(seq)

    def get_value(self) -> str:
        return self.keySequence().toString()

    def is_valid(self) -> bool:
        return True


if __name__ == "__main__":
    app = widgets.app()
    widget = KeySequenceEdit()
    widget.show()
    app.main_loop()



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtCore
from prettyqt.utils import bidict


CLOCK_TYPE = bidict(
    system_time=QtCore.QElapsedTimer.SystemTime,
    monotonic_clock=QtCore.QElapsedTimer.MonotonicClock,
    tick_counter=QtCore.QElapsedTimer.TickCounter,
    mach_absolute_time=QtCore.QElapsedTimer.MachAbsoluteTime,
    performance_counter=QtCore.QElapsedTimer.PerformanceCounter,
)

ClockTypeStr = Literal[
    "system_time",
    "monotonic_clock",
    "tick_counter",
    "mach_absolute_time",
    "performance_counter",
]


class ElapsedTimer(QtCore.QElapsedTimer):
    def __bool__(self):
        return self.isValid()

    def get_clock_type(self) -> ClockTypeStr:
        """Return current clock type.

        Returns:
            clock type
        """
        return CLOCK_TYPE.inverse[self.clockType()]



from __future__ import annotations

from prettyqt.qt import QtWidgets


class UndoCommand(QtWidgets.QUndoCommand):
    def __len__(self) -> int:
        return self.childCount()

    def __getitem__(self, index: int) -> QtWidgets.QUndoCommand:
        child = self.child(index)
        if child is None:
            raise KeyError(index)
        return child

    def serialize_fields(self):
        return dict(children=[self.child(i) for i in range(self.childCount())])

    def __setstate__(self, state):
        for c in state["children"]:
            c.setParent(self)

    def __reduce__(self):
        return type(self), (), self.__getstate__()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QItemDelegate.__bases__ = (widgets.AbstractItemDelegate,)


class ItemDelegate(QtWidgets.QItemDelegate):
    pass



from __future__ import annotations

from typing import Literal, Optional

from prettyqt import core, gui
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import InvalidParamError, bidict


MODES = bidict(
    clipboard=QtGui.QClipboard.Clipboard,
    selection=QtGui.QClipboard.Selection,
    find_buffer=QtGui.QClipboard.FindBuffer,
)

ModeStr = Literal["clipboard", "selection", "find_buffer"]


QtGui.QClipboard.__bases__ = (core.Object,)


class Clipboard:
    def __init__(self, item: QtGui.QClipboard):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def set_pixmap(self, pixmap: Optional[QtGui.QPixmap], mode: ModeStr = "clipboard"):
        if mode not in MODES:
            raise InvalidParamError(mode, MODES)
        if pixmap is None:
            pixmap = QtGui.QPixmap()
        self.item.setPixmap(pixmap, MODES[mode])

    def get_pixmap(self, mode: ModeStr = "clipboard") -> Optional[gui.Pixmap]:
        if mode not in MODES:
            raise InvalidParamError(mode, MODES)
        pix = gui.Pixmap(self.item.pixmap(MODES[mode]))
        if pix.isNull():
            return None
        return pix

    def set_image(self, image: Optional[QtGui.QImage], mode: ModeStr = "clipboard"):
        if mode not in MODES:
            raise InvalidParamError(mode, MODES)
        if image is None:
            image = QtGui.QImage()
        self.item.setImage(image, MODES[mode])

    def get_image(self, mode: ModeStr = "clipboard") -> Optional[gui.Image]:
        if mode not in MODES:
            raise InvalidParamError(mode, MODES)
        img = gui.Image(self.item.image(MODES[mode]))
        if img.isNull():
            return None
        return img

    def set_mimedata(self, mimedata: QtCore.QMimeData, mode: ModeStr = "clipboard"):
        if mode not in MODES:
            raise InvalidParamError(mode, MODES)
        self.item.setMimeData(mimedata, MODES[mode])

    def get_mimedata(self, mode: ModeStr = "clipboard") -> QtCore.QMimeData:
        if mode not in MODES:
            raise InvalidParamError(mode, MODES)
        return self.item.mimeData(MODES[mode])

    def set_text(self, text: str, mode: ModeStr = "clipboard"):
        if mode not in MODES:
            raise InvalidParamError(mode, MODES)
        self.item.setText(text, MODES[mode])

    def get_text(self, mode: ModeStr = "clipboard") -> str:
        if mode not in MODES:
            raise InvalidParamError(mode, MODES)
        return self.item.text(MODES[mode])



from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from prettyqt import constants, core, iconprovider, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, helpers, types


QtWidgets.QToolBar.__bases__ = (widgets.Widget,)


class ToolBar(QtWidgets.QToolBar):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_icon_size(24)
        self.menu_buttons = list()

    def __setstate__(self, state: Dict[str, Any]) -> None:
        super().__setstate__(state)
        self.addActions(state["actions"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def serialize_fields(self):
        return dict(actions=self.actions(), allowed_areas=self.get_allowed_areas())

    def __repr__(self):
        return f"{type(self).__name__}({self.windowTitle()!r})"

    def add(self, item: Union[QtWidgets.QAction, QtWidgets.QWidget]):
        if isinstance(item, QtWidgets.QAction):
            self.addAction(item)
        else:
            self.addWidget(item)

    def add_menu_button(
        self, label: str, icon: types.IconType, menu: QtWidgets.QMenu
    ) -> widgets.ToolButton:
        btn = widgets.ToolButton.for_menu(menu)
        btn.setText(label)
        btn.setToolButtonStyle(self.toolButtonStyle())
        btn.set_icon(icon)
        self.menu_buttons.append(btn)
        self.addWidget(btn)
        return btn

    def add_separator(
        self, text: Optional[str] = None, before: QtWidgets.QAction = None
    ) -> QtWidgets.QAction:
        """Adds a separator showing an optional label.

        Args:
            text: Text to show on separator
            before: insert separator before specific action

        Returns:
            Separator action
        """
        if text is None:
            if before:
                return self.insertSeparator(before)
            else:
                return self.addSeparator()
        else:
            label = widgets.Label(text)
            label.setMinimumWidth(self.minimumWidth())
            with label.edit_stylesheet() as ss:
                ss.background.setValue("lightgrey")
            label.set_alignment(horizontal="center")
            if before:
                return self.insertWidget(before, label)
            else:
                return self.addWidget(label)

    def set_style(self, style: constants.ToolButtonStyleStr):
        self.setToolButtonStyle(constants.TOOLBUTTON_STYLE[style])
        for btn in self.menu_buttons:
            btn.set_style(style)

    def get_style(self) -> constants.ToolButtonStyleStr:
        """Return current style.

        Returns:
            style
        """
        return constants.TOOLBUTTON_STYLE.inverse[self.toolButtonStyle()]

    def add_action(
        self,
        label: str,
        icon: types.IconType = None,
        callback: Optional[Callable] = None,
        checkable: bool = False,
    ) -> QtWidgets.QAction:
        icon = iconprovider.get_icon(icon)
        action = self.addAction(icon, label)
        if callback is not None:
            action.triggered.connect(callback)
        if checkable:
            action.setCheckable(True)
        return action

    def add_spacer(self) -> QtWidgets.QAction:
        spacer = widgets.Widget()
        spacer.set_size_policy("expanding", "expanding")
        return self.addWidget(spacer)

    def set_icon_size(self, size: Union[int, Tuple[int, int], QtCore.QSize]):
        """Set size of the icons."""
        if isinstance(size, int):
            size = core.Size(size, size)
        elif isinstance(size, tuple):
            size = core.Size(*size)
        self.setIconSize(size)

    def get_icon_size(self) -> core.Size:
        return core.Size(self.iconSize())

    def set_font_size(self, size: int):
        with self.edit_font() as font:
            font.set_size(size)

    def is_area_allowed(self, area: constants.ToolbarAreaStr) -> bool:
        """Check if toolbar is allowed at specified area.

        Args:
            area: area of the toolbar

        Raises:
            InvalidParamError: area does not exist
        """
        if area not in constants.TOOLBAR_AREA:
            raise InvalidParamError(area, constants.TOOLBAR_AREA)
        return self.isAreaAllowed(constants.TOOLBAR_AREA[area])

    def set_allowed_areas(self, *areas: constants.ToolbarAreaStr):
        for area in areas:
            if area not in constants.TOOLBAR_AREA:
                raise InvalidParamError(area, constants.TOOLBAR_AREA)
        flag = helpers.merge_flags(areas, constants.TOOLBAR_AREA)
        self.setAllowedAreas(flag)

    def get_allowed_areas(self) -> List[constants.ToolbarAreaStr]:
        return [
            k
            for k, v in constants.TOOLBAR_AREA.items()
            if v & self.allowedAreas()  # type: ignore
        ]


if __name__ == "__main__":
    app = widgets.app()
    widget = ToolBar("test")
    widget.show()
    app.main_loop()



"""Provides QtWidgets classes and functions."""

from prettyqt.qt import PYQT5, PYSIDE2, PYSIDE6, PythonQtError


if PYQT5:
    from PyQt5.QtWidgets import *
elif PYSIDE2:
    from PySide2.QtWidgets import *
elif PYSIDE6:
    from PySide6.QtWidgets import *  # type: ignore
    from PySide6.QtGui import (  # type: ignore
        QAction,
        QActionGroup,
        QUndoCommand,
        QUndoStack,
        QUndoGroup,
        QShortcut,
    )
else:
    raise PythonQtError("No Qt bindings could be found")



"""Provides QtTest classes and functions."""

from prettyqt.qt import PYQT5, PYSIDE2, PYSIDE6, PythonQtError


if PYQT5:
    from PyQt5 import uic

    class QUiLoader:
        def load(self, path: str):
            return uic.loadUi(path)

    RCC_CMD = "pyrcc5 -no-compress"
    LUPDATE_CMD = "pylupdate5 -verbose"
    UIC_CMD = "pyuic5 --debug"
elif PYSIDE2:
    from PySide2.QtUiTools import QUiLoader, loadUiType

    RCC_CMD = "pyside2-rcc --no-compress --verbose"
    LUPDATE_CMD = "pyside2-lupdate -verbose"
    UIC_CMD = "pyside2-uic"
elif PYSIDE6:
    from PySide6.QtUiTools import QUiLoader, loadUiType  # type: ignore

    RCC_CMD = "pyside6-rcc --no-compress --verbose"
    LUPDATE_CMD = "pyside6-lupdate -verbose"
    UIC_CMD = "pyside6-uic"
else:
    raise PythonQtError("No Qt bindings could be found")



from __future__ import annotations

from collections import defaultdict
import contextlib
import inspect
import itertools
from typing import Any, DefaultDict, Dict, List, Optional, Type, TypeVar, Union

from prettyqt import core
import prettyqt.qt
from prettyqt.qt import QtCore


T = TypeVar("T", bound=QtCore.QObject)

counter_dict: DefaultDict = defaultdict(itertools.count)


class Object(QtCore.QObject):
    def __repr__(self):
        return f"{type(self).__name__}()"

    def __setstate__(self, state):
        self.set_id(state["object_name"])

    def __getstate__(self):
        return self.serialize()

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def serialize_fields(self):
        return dict(object_name=self.objectName())

    def serialize(self) -> Dict[str, Any]:
        dct = {}
        for klass in reversed(inspect.getmro(type(self))):
            if "serialize_fields" in klass.__dict__:
                data = klass.serialize_fields(self)  # type: ignore
                dct.update(data)
        return dct

    @contextlib.contextmanager
    def block_signals(self):
        self.blockSignals(True)
        yield None
        self.blockSignals(False)

    def to_json(self):
        dct = self.__getstate__()
        for k, v in dct.items():
            if isinstance(v, QtCore.QObject):
                dct[k] = v.to_json()
        return dct

    def set_unique_id(self):
        class_name = type(self).__name__
        count = next(counter_dict[class_name])
        self.set_id(f"{class_name}_{count}")

    def set_id(self, name: str):
        self.setObjectName(name)

    def get_id(self) -> str:
        return self.objectName()

    def has_id(self) -> bool:
        return self.objectName() != ""

    @classmethod
    def get_metaobject(cls) -> core.MetaObject:
        return core.MetaObject(cls.staticMetaObject)

    # @property
    # def id(self) -> str:
    #     return self.objectName()

    # @id.setter
    # def id(self, name: str):
    #     self.setObjectName(name)

    def find_children(
        self,
        typ: Type[T] = QtCore.QObject,
        name: Optional[Union[str, QtCore.QRegularExpression]] = None,
        recursive: bool = True,
    ) -> List[T]:
        if recursive:
            flag = QtCore.Qt.FindChildrenRecursively
        else:
            flag = QtCore.Qt.FindDirectChildrenOnly
        if prettyqt.qt.API == "pyqt5":
            return self.findChildren(typ, name=name, options=flag)  # type: ignore
        else:
            if name is None:
                items = self.findChildren(typ)
            else:
                items = self.findChildren(typ, name)
            return [i for i in items if recursive or i.parent() == self]

    def find_child(
        self,
        typ: Type[T] = QtCore.QObject,
        name: Optional[Union[str, QtCore.QRegularExpression]] = None,
        recursive: bool = True,
    ) -> Optional[T]:
        if recursive:
            flag = QtCore.Qt.FindChildrenRecursively
        else:
            flag = QtCore.Qt.FindDirectChildrenOnly
        if prettyqt.qt.API == "pyqt5":
            return self.findChild(typ, name, flag)  # type: ignore
        else:
            if name is None:
                item = self.findChild(typ)
            else:
                item = self.findChild(typ, name)
            return item if recursive or item.parent() == self else None  # type: ignore

    def find_parent(
        self, typ: Type[QtCore.QObject], name: Optional[str] = None
    ) -> Optional[QtCore.QObject]:
        node = self
        while node:
            node = node.parent()
            if isinstance(node, typ):
                if name is None or node.objectName() == name:
                    return node
        return None



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QFocusFrame.__bases__ = (widgets.Widget,)


class FocusFrame(QtWidgets.QFocusFrame):
    pass


if __name__ == "__main__":
    app = widgets.app()
    widget = FocusFrame()
    widget.show()
    app.main_loop()



from __future__ import annotations

import regex as re

from prettyqt import gui


BRACKETS = re.compile(r"(\{|\}|\[|\]|\:|\,)")
REGEXP1 = re.compile(r"\".*\" *\:")
REGEXP2 = re.compile(r"\: *\".*\"")

SYMBOL_FORMAT = gui.TextCharFormat()
SYMBOL_FORMAT.set_foreground_color("red")
SYMBOL_FORMAT.set_font_weight("bold")

NAME_FORMAT = gui.TextCharFormat()
NAME_FORMAT.set_foreground_color("blue")

VALUE_FORMAT = gui.TextCharFormat()
VALUE_FORMAT.set_foreground_color("darkgreen")


class JsonHighlighter(gui.SyntaxHighlighter):
    def highlightBlock(self, text: str):
        """Highlight a block of code using the rules outlined in the Constructor."""
        for m in BRACKETS.finditer(text):
            self.setFormat(m.span()[0], m.span()[1] - m.span()[0], SYMBOL_FORMAT)

        text.replace('\\"', "  ")
        for m in REGEXP1.finditer(text):
            self.setFormat(m.span()[0], m.span()[1] - m.span()[0], NAME_FORMAT)
        for m in REGEXP2.finditer(text):
            self.setFormat(m.span()[0], m.span()[1] - m.span()[0], VALUE_FORMAT)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    editor = widgets.PlainTextEdit()
    highlighter = JsonHighlighter(editor.document())
    editor.show()
    app.main_loop()



"""A lightweight module handling iconic fonts.

It is designed to provide a simple way for creating QtGui.QIcons from glyphs.

From a user's viewpoint, the main entry point is the ``IconicFont`` class which
contains methods for loading new iconic fonts with their character map and
methods returning instances of ``QIcon``.

"""

from __future__ import annotations

# Standard library imports
from typing import Any, Dict, List

from prettyqt import constants, core, gui
from prettyqt.qt import QtCore, QtGui


# Third party imports

_default_options = {
    "color": gui.Color(50, 50, 50),
    "color_disabled": gui.Color(150, 150, 150),
    "opacity": 1.0,
    "scale_factor": 1.0,
}

ZERO_COORD = core.Point(0, 0)

ICON_KW = [
    "char",
    "on",
    "off",
    "active",
    "selected",
    "disabled",
    "on_active",
    "on_selected",
    "on_disabled",
    "off_active",
    "off_selected",
    "off_disabled",
]

VALID_OPTIONS = [
    "active",
    "selected",
    "disabled",
    "on",
    "off",
    "on_active",
    "on_selected",
    "on_disabled",
    "off_active",
    "off_selected",
    "off_disabled",
    "color",
    "color_on",
    "color_off",
    "color_active",
    "color_selected",
    "color_disabled",
    "color_on_selected",
    "color_on_active",
    "color_on_disabled",
    "color_off_selected",
    "color_off_active",
    "color_off_disabled",
    "animation",
    "offset",
    "scale_factor",
    "rotated",
    "hflip",
    "vflip",
]

COLOR_OPTIONS = {
    QtGui.QIcon.On: {
        QtGui.QIcon.Normal: ("color_on", "on"),
        QtGui.QIcon.Disabled: ("color_on_disabled", "on_disabled"),
        QtGui.QIcon.Active: ("color_on_active", "on_active"),
        QtGui.QIcon.Selected: ("color_on_selected", "on_selected"),
    },
    QtGui.QIcon.Off: {
        QtGui.QIcon.Normal: ("color_off", "off"),
        QtGui.QIcon.Disabled: ("color_off_disabled", "off_disabled"),
        QtGui.QIcon.Active: ("color_off_active", "off_active"),
        QtGui.QIcon.Selected: ("color_off_selected", "off_selected"),
    },
}


def set_global_defaults(**kwargs):
    """Set global defaults for the options passed to the icon painter."""
    for kw in kwargs:
        if kw not in VALID_OPTIONS:
            raise KeyError(f"Invalid option {kw!r}")
        _default_options[kw] = kwargs[kw]


class CharIconPainter:
    """Char icon painter."""

    def paint(
        self,
        iconic: IconicFont,
        painter: gui.Painter,
        rect,
        mode,
        state,
        options: List[Dict[str, Any]],
    ):
        color_str, char = COLOR_OPTIONS[state][mode]
        for opt in options:
            painter.save()
            color = gui.Color(opt[color_str])
            painter.setPen(color)

            # A 16 pixel-high icon yields a font size of 14, which is pixel perfect
            # for font-awesome. 16 * 0.875 = 14
            # The reason why the glyph size is smaller than the icon size is to
            # account for font bearing.

            draw_size = round(0.875 * rect.height() * opt["scale_factor"])
            # Animation setup hook
            if (animation := opt.get("animation")) is not None:
                animation.setup(self, painter, rect)
            font = iconic.fonts[opt["prefix"]].get_font(draw_size)
            painter.setFont(font)
            if "offset" in opt:
                rect.translate(
                    round(opt["offset"][0] * rect.width()),
                    round(opt["offset"][1] * rect.height()),
                )
            x_center = rect.width() * 0.5
            y_center = rect.height() * 0.5
            painter.translate(x_center, y_center)
            transform = gui.Transform()
            if opt.get("vflip") is True:
                transform.scale(1, -1)
            if opt.get("hflip") is True:
                transform.scale(-1, 1)
            if "rotated" in opt:
                transform.rotate(opt["rotated"])
            painter.setTransform(transform, True)
            painter.translate(-x_center, -y_center)
            if (opacity := opt.get("opacity")) is not None:
                painter.setOpacity(opacity)

            painter.drawText(rect, int(constants.ALIGN_CENTER), opt[char])
            painter.restore()


class FontError(Exception):
    """Exception for font errors."""


class CharIconEngine(gui.IconEngine):
    """Specialization of QtGui.QIconEngine used to draw font-based icons."""

    def __init__(self, iconic: QtCore.QObject, painter, options):
        super().__init__()
        self.iconic = iconic
        self.painter = painter
        self.options = options

    def paint(self, painter: QtGui.QPainter, rect: QtCore.QRect, mode, state):
        self.painter.paint(self.iconic, painter, rect, mode, state, self.options)

    def pixmap(self, size, mode, state) -> QtGui.QPixmap:
        pm = QtGui.QPixmap(size)
        pm.fill(QtCore.Qt.transparent)
        rect = core.Rect(ZERO_COORD, size)
        painter = gui.Painter(pm)
        self.paint(painter, rect, mode, state)
        return pm


class IconicFont(core.Object):
    """Main class for managing iconic fonts."""

    def __init__(self, *args):
        """Initialize IconicFont.

        Parameters
        ----------
        ``*args``: tuples
            Each positional argument is a tuple of 3 or 4 values:
            - The prefix string to be used when accessing a given font set,
            - The ttf font filename,
            - The json charmap filename,
            - Optionally, the directory containing these files. When not
              provided, the files will be looked for in ``./fonts/``.
        """
        super().__init__()
        self.painter = CharIconPainter()
        self.icon_cache = {}
        self.fonts = {font.prefix: font for font in args}

    def has_valid_font_ids(self) -> bool:
        """Validate instance's font ids are loaded to QFontDatabase.

        It is possible that QFontDatabase was reset or QApplication was recreated
        in both cases it is possible that font is not available.
        """
        # Check stored font ids are still available
        return all(font.is_valid() for font in self.fonts.values())

    def icon(self, *names, **kwargs) -> QtGui.QIcon:
        """Return a QtGui.QIcon object corresponding to the provided icon name."""
        cache_key = f"{names}{kwargs}"
        if cache_key in self.icon_cache:
            return self.icon_cache[cache_key]
        opts = kwargs.pop("options", [{}] * len(names))
        if len(opts) != len(names):
            raise TypeError(f'"options" must be a list of size {len(names)}')
        parsed_options = [self._parse_options(o, kwargs, n) for o, n in zip(opts, names)]
        engine = CharIconEngine(self, self.painter, parsed_options)
        icon = QtGui.QIcon(engine)
        self.icon_cache[cache_key] = icon
        return icon

    def _parse_options(
        self, specific_options: dict, general_options: dict, name: str
    ) -> Dict[str, Any]:
        options = dict(_default_options, **general_options)
        options.update(specific_options)

        # Handle icons for modes (Active, Disabled, Selected, Normal)
        # and states (On, Off)
        char = options.get("char", name)
        on = options.get("on", char)
        off = options.get("off", char)
        active = options.get("active", on)
        selected = options.get("selected", active)
        disabled = options.get("disabled", char)
        on_active = options.get("on_active", active)
        on_selected = options.get("on_selected", selected)
        on_disabled = options.get("on_disabled", disabled)
        off_active = options.get("off_active", active)
        off_selected = options.get("off_selected", selected)
        off_disabled = options.get("off_disabled", disabled)

        icon_dict = {
            "char": char,
            "on": on,
            "off": off,
            "active": active,
            "selected": selected,
            "disabled": disabled,
            "on_active": on_active,
            "on_selected": on_selected,
            "on_disabled": on_disabled,
            "off_active": off_active,
            "off_selected": off_selected,
            "off_disabled": off_disabled,
        }
        names = [icon_dict.get(kw, name) for kw in ICON_KW]
        chars = []
        for name in names:
            if "." not in name:
                raise Exception("Invalid icon name")
            prefix, n = name.split(".")
            if prefix not in self.fonts:
                raise Exception(f"Invalid font prefix {prefix!r}")
            if n not in self.fonts[prefix].charmap:
                raise Exception(f"Invalid icon name {n!r} in font {prefix!r}")
            chars.append(self.fonts[prefix].charmap[n])
        options.update(dict(zip(*(ICON_KW, chars))))
        options.update({"prefix": prefix})

        # Handle colors for modes (Active, Disabled, Selected, Normal)
        # and states (On, Off)
        color = options.get("color")
        options.setdefault("color_on", color)
        options.setdefault("color_active", options["color_on"])
        options.setdefault("color_selected", options["color_active"])
        options.setdefault("color_on_active", options["color_active"])
        options.setdefault("color_on_selected", options["color_selected"])
        options.setdefault("color_on_disabled", options["color_disabled"])
        options.setdefault("color_off", color)
        options.setdefault("color_off_active", options["color_active"])
        options.setdefault("color_off_selected", options["color_selected"])
        options.setdefault("color_off_disabled", options["color_disabled"])
        return options



from __future__ import annotations

from typing import Any, Literal

from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError, bidict


QScrollerProperties = QtWidgets.QScrollerProperties

FRAME_RATES = bidict(
    fps_60=QScrollerProperties.Fps60,
    fps_30=QScrollerProperties.Fps30,
    fps_20=QScrollerProperties.Fps20,
    standard=QScrollerProperties.Standard,
)

FrameRateStr = Literal["fps_60", "fps_30", "fps_20", "standard"]

OVERSHOOT_POLICY = bidict(
    when_scrollable=QScrollerProperties.OvershootWhenScrollable,
    always_off=QScrollerProperties.OvershootAlwaysOff,
    always_on=QScrollerProperties.OvershootAlwaysOn,
)

OvershootPolicyStr = Literal["when_scrollable", "always_off", "always_on"]

SCROLL_METRIC = bidict(
    mouse_press_event_delay=QScrollerProperties.MousePressEventDelay,
    drag_start_distance=QScrollerProperties.DragStartDistance,
    drag_velocity_smoothing_factor=QScrollerProperties.DragVelocitySmoothingFactor,
    axis_lock_threshold=QScrollerProperties.AxisLockThreshold,
    scrolling_curve=QScrollerProperties.ScrollingCurve,
    deceleration_factor=QScrollerProperties.DecelerationFactor,
    minimum_velocity=QScrollerProperties.MinimumVelocity,
    maximum_velocity=QScrollerProperties.MaximumVelocity,
    maximum_click_through_velocity=QScrollerProperties.MaximumClickThroughVelocity,
    accelerating_flick_maximum_time=QScrollerProperties.AcceleratingFlickMaximumTime,
    accelerating_flick_speedup_factor=QScrollerProperties.AcceleratingFlickSpeedupFactor,
    snap_position_ratio=QScrollerProperties.SnapPositionRatio,
    snap_time=QScrollerProperties.SnapTime,
    overshoot_drag_resistance_factor=QScrollerProperties.OvershootDragResistanceFactor,
    overshoot_drag_distance_factor=QScrollerProperties.OvershootDragDistanceFactor,
    overshoot_scroll_distance_factor=QScrollerProperties.OvershootScrollDistanceFactor,
    overshoot_scroll_time=QScrollerProperties.OvershootScrollTime,
    horizontal_overshoot_policy=QScrollerProperties.HorizontalOvershootPolicy,
    vertical_overshoot_policy=QScrollerProperties.VerticalOvershootPolicy,
    frame_rate=QScrollerProperties.FrameRate,
    scroll_metric_count=QScrollerProperties.ScrollMetricCount,
)

ScrollmetricStr = Literal[
    "mouse_press_event_delay",
    "drag_start_distance",
    "drag_velocity_smoothing_factor",
    "axis_lock_threshold",
    "scrolling_curve",
    "deceleration_factor",
    "minimum_velocity",
    "maximum_velocity",
    "maximum_click_through_velocity",
    "accelerating_flick_maximum_time",
    "accelerating_flick_speedup_factor",
    "snap_position_ratio",
    "snap_time",
    "overshoot_drag_resistance_factor",
    "overshoot_drag_distance_factor",
    "overshoot_scroll_distance_factor",
    "overshoot_scroll_time",
    "horizontal_overshoot_policy",
    "vertical_overshoot_policy",
    "frame_rate",
    "scroll_metric_count",
]


class ScrollerProperties(QtWidgets.QScrollerProperties):
    def __getitem__(self, metric: ScrollmetricStr):
        return self.get_scroll_metric(metric)

    def __setitem__(self, metric: ScrollmetricStr, value: Any):
        self.set_scroll_metric(metric, value)

    def set_scroll_metric(self, metric: ScrollmetricStr, value: Any):
        """Set scroll metric.

        Args:
            metric: Scroll metric to set
            value: Value to set

        Raises:
            InvalidParamError: metric does not exist

        """
        if metric not in SCROLL_METRIC:
            raise InvalidParamError(metric, SCROLL_METRIC)
        self.setScrollMetric(SCROLL_METRIC[metric], value)

    def get_scroll_metric(self, metric: ScrollmetricStr) -> Any:
        """Return scroll metric.

        Args:
            metric: Scroll metric to get

        Raises:
            InvalidParamError: metric does not exist

        Returns:
            state
        """
        if metric not in SCROLL_METRIC:
            raise InvalidParamError(metric, SCROLL_METRIC)
        return self.scrollMetric(SCROLL_METRIC[metric])



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QStyleOptionComboBox.__bases__ = (widgets.StyleOptionComplex,)


class StyleOptionComboBox(QtWidgets.QStyleOptionComboBox):
    pass



from __future__ import annotations

from typing import Optional, Tuple, Union

from prettyqt import core, gui
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import InvalidParamError


QtGui.QPdfWriter.__bases__ = (core.Object, gui.PagedPaintDevice)


class PdfWriter(QtGui.QPdfWriter):
    def set_desktop_resolution(self):
        primary = gui.app().primaryScreen()
        dpi = primary.logicalDotsPerInch()
        self.setResolution(int(dpi))

    def set_page_margins(
        self,
        margins: Union[Tuple[float, float, float, float], QtCore.QMarginsF],
        unit: Optional[gui.pagelayout.UnitStr] = None,
    ) -> bool:
        if isinstance(margins, tuple):
            margins = QtCore.QMarginsF(*margins)
        if unit is None:
            return self.setPageMargins(margins)
        else:
            if unit not in gui.pagelayout.UNITS:
                raise InvalidParamError(unit, gui.pagelayout.UNITS)
            return self.setPageMargins(margins, gui.pagelayout.UNITS[unit])


if __name__ == "__main__":
    app = gui.app()
    writer = PdfWriter("")
    writer.set_desktop_resolution()



from __future__ import annotations

import os
import pathlib
from typing import Union

from prettyqt import core, gui
from prettyqt.qt import QtCore, QtGui


QtGui.QPixmap.__bases__ = (gui.PaintDevice,)


class Pixmap(QtGui.QPixmap):
    def __bool__(self):
        return not self.isNull()

    def __getstate__(self):
        return bytes(self)

    def __setstate__(self, ba):
        core.DataStream.write_bytearray(ba, self)

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __bytes__(self):
        ba = core.DataStream.create_bytearray(self)
        return bytes(ba)

    def __eq__(self, other):
        if not isinstance(other, Pixmap):
            return False
        # return bytes(self) == bytes(other)
        return self.toImage() == other.toImage()

    def __hash__(self):
        return self.cacheKey()

    @classmethod
    def from_file(cls, path: Union[os.PathLike, str]) -> Pixmap:
        path = pathlib.Path(path)
        with path.open(mode="rb") as f:
            data = f.read()
        # Create widget
        pixmap = cls()
        pixmap.loadFromData(QtCore.QByteArray(data))
        return pixmap

    def get_size(self) -> core.Size:
        return core.Size(self.size())

    def get_rect(self) -> core.Rect:
        return core.Rect(self.rect())

    def to_image(self) -> gui.Image:
        return gui.Image(self.toImage())

    @classmethod
    def create_dot(cls, color="black", size: int = 16) -> Pixmap:
        col = gui.Color(color)
        px = cls(size, size)
        px.fill(QtCore.Qt.transparent)
        px_size = px.rect().adjusted(1, 1, -1, -1)
        with gui.Painter(px) as painter:
            painter.use_antialiasing()
            painter.setBrush(col)
            painter.set_pen(color=gui.Color(15, 15, 15), width=1)
            painter.drawEllipse(px_size)
        return px


if __name__ == "__main__":
    from prettyqt import gui

    app = gui.app()
    p = Pixmap()
    print(bytes(p))



# from prettyqt.qt import QtCore

# from prettyqt import core


from __future__ import annotations


# QtCore.QRandomGenerator64.__bases__ = (core.RandomGenerator,)

# class RandomGenerator64(QtCore.QRandomGenerator64):
#     pass



from __future__ import annotations

import datetime
from typing import Union

from prettyqt import core, widgets
from prettyqt.qt import QtCore, QtWidgets


QtWidgets.QTimeEdit.__bases__ = (widgets.DateTimeEdit,)


class TimeEdit(QtWidgets.QTimeEdit):

    value_changed = core.Signal(datetime.datetime)

    def serialize_fields(self):
        return dict(
            time=self.get_time(),
            range=(self.min_time(), self.max_time()),
        )

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setTime(state["time"])
        self.set_range(*state["range"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def set_range(
        self,
        lower: Union[QtCore.QTime, datetime.time],
        upper: Union[QtCore.QTime, datetime.time],
    ):
        self.setToolTip(f"{lower} <= x <= {upper}")
        self.setTimeRange(lower, upper)  # type: ignore

    def get_value(self) -> datetime.time:
        return self.get_time()

    def set_value(self, value: Union[datetime.time, QtCore.QTime]):
        return self.setTime(value)  # type: ignore


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = TimeEdit()
    widget.show()
    app.main_loop()
